#!/usr/bin/env python
"""Sommon common plots that are not directly implemented in :mod:`matplotlib`,
as well as some specific plots used by :mod:`plastid`. Demos of these appear
in :doc:`/examples/z_plotting`.

General plots
-------------
    :func:`stacked_bar`
        Create a stacked bar graph

    :func:`kde_plot`
        Plot a kernel density estimate (continuous histogram) of data

    :func:`scatterhist_x`, :func:`scatterhist_y`, and :func:`scatterhist_xy`
        Create scatter plots with kernel density estimates of the marginal 
        distributions along side them

    :func:`profile_heatmap`
        Plot a heatmap, with a columnwise median (or other profile summarizing
        the heatmap) in an upper panel above it, with aligned coordinates

    :func:`triangle_plot`
        Plot data lying on the plane x + y + z = k (e.g. codon phasing)
        in a homogeneous 2D representation


Plots for genomics
------------------
    :func:`ma_plot`
        Plot fold changes between `x` and `y` (:math:`\log_{2} (y/x)`) as
        a function of the mean of x and y (:math:`0.5*(x+y)`).

    :func:`phase_plot`
        For ribosome profiling. Plot sub-codon phasing of ribosome-protected
        foorpints stratified by read length, as well as the fraction of total
        reads represented by each length.

"""
import copy
import numpy
import scipy.stats
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.artist import Artist
from plastid.plotting.colors import lighten, darken, process_black
from plastid.plotting.plotutils import get_fig_axes, split_axes, clean_invalid, get_kde

#==============================================================================
# Default keyword arguments for plots or subplots
# and helper functions
#==============================================================================

plastid_default_scatter = {
    "marker"    : "o",
    "alpha"     : 0.7,
    "facecolor" : "none",
    "s"         : 8,
    "rasterized": True,
} # yapf: disable
"""Default parameters for scatter plots"""


def get_color_cycle(ax):
    """Get color cycle iterator from multiple versions of matplotlib axes

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`


    Returns
    -------
    iterator
        Iterator over colors, passable to matplotlib `color` keyword
    """
    if hasattr(ax, "_get_lines"):
        if hasattr(ax._get_lines, "prop_cycler"):
            return (X["color"] for X in ax._get_lines.prop_cycler)
        elif hasattr(ax._get_lines, "color_cycle"):
            return ax._get_lines.color_cycle
    else:
        raise AssertionError("get_color_cycle: Could not find color cycle")


#==============================================================================
# Stacked bar
#==============================================================================


def stacked_bar(data, axes=None, labels=None, lighten_by=0.1, cmap=None, **kwargs):
    """Create a stacked bar graph
    
    Parameters
    ----------
    data : :class:`numpy.ndarray`
        Array of data, in which each row is a stack, each column a value in that stack.

    axes : :class:`matplotlib.axes.Axes` or `None`, optional
        Axes in which to place plot. If `None`, a new figure is generated.
        (Default: `None`)
        
    labels : list, optional
        Labels for each stack. If `None`, stacks are labeled sequentially by number.
        (Default: `None`)
    
    lighten_by : float, optional
        Amount by which to lighten sequential blocks in each stack. (Default: 0.10)
    
    cmap : :class:`matplotlib.colors.Colormap`, optional
        Colormap from which to generate bar colors. If supplied, will override
        any `color` attribute in `**kwargs`. (Default: `None`)
        
    **kwargs : keyword arguments
        Other keyword arguments to pass to :func:`matplotlib.pyplot.bar`

    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Parent figure of axes
    
    :class:`matplotlib.axes.Axes`
        Axes containing plot
    """
    fig, ax = get_fig_axes(axes)
    rows, cols = data.shape
    labels = labels if labels is not None else range(rows)
    defaults = [("align", "center"), ("width", 0.8)]

    if cmap is not None:
        kwargs["color"] = cmap(numpy.linspace(0, 1.0, num=10))
    elif kwargs.get("color", None) is None:
        kwargs["color"] = [next(get_color_cycle(ax)) for _ in range(rows)]

    x = numpy.arange(rows) + 0.5
    xaxis = ax.xaxis
    xaxis.set_ticks(x)
    xaxis.set_ticklabels(labels)
    bottoms = numpy.zeros(rows)

    for k, v in defaults:
        if k not in kwargs:
            kwargs[k] = v

    for i in range(cols):
        color = kwargs["color"]
        if i > 0:
            kwargs["color"] = lighten(color, amt=lighten_by)

        heights = data[:, i]
        plt.bar(x, heights, bottom=bottoms, **kwargs)
        heights.shape
        bottoms += heights

    ax.set_xlim(-0.5, rows + 0.5)

    return fig, ax


#==============================================================================
# Kernel density estimate
#==============================================================================

def kde_plot(
        data,
        axes      = None,
        color     = None,
        label     = None,
        alpha     = 0.7,
        vert      = False,
        log       = False,
        base      = 10,
        points    = 500,
        bw_method = "scott",
        rescale   = False,
        zorder    = None,
        fill      = True
): # yapf: disable
    """Plot a kernel density estimate of `data` on `axes`.

    Parameters
    ----------
    data : :class:`numpy.ndarray`
        Array of data

    axes : :class:`matplotlib.axes.Axes` or `None`, optional
        Axes in which to place plot. If `None`, a new figure is generated.
        (Default: `None`)
        
    color : matplotlib colorspec, optional
        Color to use for plotting (Default: use next in matplotlibrc)

    label : str, optional
        Name of data series (used for legend; default: `None`)

    alpha : float, optional
        Amount of alpha transparency to use (Default: 0.7)

    vert : bool, optional
        If true, plot kde vertically

    log : bool, optional
        If `True`, `data` is log-transformed before the kde is estimated.
        Data are converted back to non-log space afterwards.

    base : 2, 10, or :obj:`numpy.e`, optional
        If `log` is `True`, this serves as the base of the log space.
        If `log` is `False`, this is ignored. (Default: 2)

    points : int
        Number of points over which to evaluate kde. (Default: 100)

    bw_method : str
        Bandwith estimation method. See documentation for
        :obj:`scipy.stats.gaussian_kde`. (Default: "scott")    


    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Parent figure of axes
    
    :class:`matplotlib.axes.Axes`
        Axes containing plot
    """
    fig, axes = get_fig_axes(axes)

    if color is None:
        color = next(get_color_cycle(axes))

    a, b = get_kde(data, log=log, base=base, points=points, bw_method=bw_method)

    if rescale == True:
        b /= b.max()

    fbargs = {"alpha": alpha, "facecolor": lighten(color), "edgecolor": color}
    if label is not None:
        fbargs["label"] = label

    if vert == True:
        if fill == True:
            axes.fill_betweenx(a, b, 0, **fbargs)

        # plot twice because in some versions of matplotlib, 
        # fill_between doesn't add to legend
        # TODO: remove in future when mpl's behavior changes
        axes.plot(
            b, a, color=color, alpha=alpha, label=label
        )
        if log == True:
            axes.semilogy()
    else:
        if fill == True:
            axes.fill_between(a, b, 0, **fbargs)
        axes.plot(a, b, color=color, alpha=alpha, label=label)
        if log == True:
            axes.semilogx()

    return fig, axes


#==============================================================================
# Triangle plot
#==============================================================================

rotate = numpy.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]])

_triA = numpy.array([[1, 0, 0], [-1, -1, 1]])
_triT = numpy.array([[0.5, 1], [0.5 * (3**0.5), 0]])
_triTA = _triT.dot(_triA)

_triverts = numpy.array([[1.0, 0.0], [0.0, 1.0], [0.0, 0.0]])


def trianglize(data):
    """Convert points from triangular space to Cartesian space for plotting.
    Called internally by :func:`triangle_plot`.

    Parameters
    ----------
    data : class:`numpy.ndarray`
        Mx2 or Mx3 list or array of points in triangular space, where
        the first column is the first coordinate, the second column
        the second, and the third, the third.
    
    Returns
    -------
    :class:`numpy.ndarray`
        Corresponding Nx2 array of points in Cartesian space, for plotting
        on a standard axis
    """

    if data.shape[1] == 2:
        data = _triTA.dot(numpy.hstack([data, numpy.ones((data.shape[0], 1))]).T).T
    else:
        data = _triT.dot(data[:, (0, 2)].T).T

    return data


def triangle_plot(
        data,
        axes=None,
        fn="scatter",
        vertex_labels=None,
        grid=None,
        clip=True,
        do_setup=True,
        **kwargs
):
    """Plot data lying in a plane x + y + z = k in a homogenous triangular space.

    Parameters
    ----------
    data : :class:`numpy.ndarray`
        Mx2 or Mx3 list or array of points in triangular space, where
        the first column is the first coordinate, the second column
        the second, and the third, the third.

    axes : :class:`matplotlib.axes.Axes` or `None`, optional
        Axes in which to place plot. If `None`, a new figure is generated.
        (Default: `None`)
       
    fn : str, optional
        Name of plotting function. Must correspond to an attribute of a
        :class:`matplotlib.axes.Axes` (e.g. 'scatter', 'plot', 'hexbin'et c.),
        that is be able to take an Nx2 :class:`numpy.ndarray` in Cartesian space
        as input (e.g. 'plot', 'scatter', 'hexbin'; Default: 'scatter').

    vertex_labels : list or None, optional
        Labels for vertex. If `None`, vertices aren't labeled. (Default: `None`)

    grid : :class:`numpy.ndarray` or None, optional
        If not `None`, draw gridlines at intervals specified in `grid`,
        as long as the grid coordinate is > 0.33333 (center of triangle)
        and <= 1.0 (border).

    clip : bool, optional
        If `True` clipping masks corresponding to the triangle boundaries
        will be applied to all plot elements (Default: `True`)
        
    do_setup : bool, optional
        If `True`, the plot area will be prepared. A triangle will be drawn,
        gridlines drawn, et c. Specify `False` if plotting a second dataset
        ontop of an already-prepared axes (Default: `True`)

    **kwargs : keyword arguments
        Other keyword arguments to pass to function specified by `fn`.

    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Parent figure of axes
    
    :class:`matplotlib.axes.Axes`
        Axes containing plot

    """
    fig, ax = get_fig_axes(axes)
    mplrc = matplotlib.rcParams

    if do_setup == True:
        triverts = trianglize(_triverts)
        tripatch = matplotlib.patches.Polygon(
            triverts,
            closed=True,
            facecolor=mplrc["axes.facecolor"],
            edgecolor=mplrc["axes.edgecolor"],
            linewidth=mplrc["axes.linewidth"],
            zorder=-10
        )
        ax.add_patch(tripatch)

        # format axes
        ax.set_xlim((0, 1))
        ax.set_ylim((0, _triverts[:, 1].max()))
        ax.set_frame_on(False)
        ax.set_xticks([])
        ax.set_yticks([])

        # label vertices
        if vertex_labels is not None:
            l1, l2, l3 = vertex_labels

            tkwargs = {"fig": fig, "units": "points"}
            p1trans = matplotlib.transforms.offset_copy(ax.transData, x=0, y=8, **tkwargs)
            p2trans = matplotlib.transforms.offset_copy(ax.transData, x=-10, y=-12, **tkwargs)
            p3trans = matplotlib.transforms.offset_copy(ax.transData, x=10, y=-12, **tkwargs)
            ax.text(triverts[0, 0], triverts[0, 1], l1, transform=p1trans)
            ax.text(triverts[1, 0], triverts[1, 1], l2, transform=p2trans)
            ax.text(triverts[2, 0], triverts[2, 1], l3, transform=p3trans)

        # add gridlines
        grid_kwargs = {
            K.replace("grid.", ""): V
            for (K, V) in mplrc.items() if K.startswith("grid")
        }
        if grid is not None:
            grid = numpy.array(grid)
            remainders = (1.0 - grid) / 2
            for i, r in zip(grid, remainders):
                if i >= 1.0 / 3:
                    points = [numpy.array([i, r, r])]
                    for _ in range(3):
                        points.append(rotate.dot(points[-1]))

                    points = numpy.array(points)
                    points = trianglize(points[:, [0, 2]])

                    myline = matplotlib.lines.Line2D(points[:, 0], points[:, 1], **grid_kwargs)
                    ax.add_line(myline)

    # scale data
    data = trianglize(data)

    # plot data
    artists = []
    fn = getattr(ax, fn)
    res = fn(*zip(*data), **kwargs)
    if isinstance(res, Artist):
        artists.append(res)
    elif isinstance(res, list):
        artists.extend([X for X in res if isinstance(X, Artist)])

    # clip
    if clip == True:
        for artist in artists:
            artist.set_clip_path(tripatch)
            artist.set_clip_on(True)

    return fig, ax


#==============================================================================
# Heatmaps with profiles on top
#==============================================================================


def sort_max_position(data):
    """Generate indices that sort rows in `data` by column in which
    the row's maximal value is attained

    Parameters
    ----------
    data : :class:`numpy.ndarray`

    Returns
    -------
    :class:`numpy.ndarray`
        Indices of rows that sort data by max position
    """
    maxvals = numpy.nanmax(data, 1)
    maxidx = numpy.zeros(len(maxvals))
    for i, maxval in enumerate(maxvals):
        maxidx[i] = (data[i, :] == maxval).argmax()

    return numpy.argsort(maxidx)


_heatmap_defaults = {
    "aspect": "auto",
    "origin": "upper",
    "interpolation": "none",
}

def profile_heatmap(
        data,
        profile   = None,
        x         = None,
        axes      = None,
        sort_fn   = sort_max_position,
        cmap      = None,
        nancolor  = "#666666",
        im_args   = {},
        plot_args = {}
): # yapf: disable
    """Create a dual-paned plot in which `profile` is displayed in a top
    panel, above a heatmap showing the intensities of each row of `data`,
    optionally sorted top-to-bottom by `sort_fn`.

    Parameters
    ----------
    data : :class:`numpy.ndarray`
        Array of data, in which each row is an individual aligned vector of data
        for region of interest, and each column a position in that vector

    profile : :class:`numpy.ndarray` or None
        Reduced profile of data, often a column-wise median. If not
        supplied, it will be calculated.

    x : :class:`numpy.ndarray`
        Array of values for X-axis

    axes : :class:`matplotlib.axes.Axes` or `None`, optional
        Axes in which to place plot. If `None`, a new figure is generated.
        (Default: `None`)
       
    sort_fn : function, optional
        Sort rows in `data` by this function before plotting. Function must
        return a :class:`numpy.ndarray` of indices corresponding to rows in `data`
        (Default: sort by ascending argmax of each row)

    cmap : :class:`~matplotlib.colors.Colormap`, optional
        Colormap to use in heatmap. It not `None`, overrides any value
        in `im_args`. (Default: `None`) 

    nancolor : str or matplotlib colorspec
        Color used for plotting `nan` and other illegal or masked values
        
    im_args : dict
        Keyword arguments to pass to :func:`matplotlib.pyplot.imshow`

    plot_args : dict
        Keyword arguments to pass to :func:`matplotlib.pyplot.plot`
        for plotting metagene average


    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Parent figure of axes
    
    dict
        Dictionary of :class:`matplotlib.axes.Axes`. "top" refers to the 
        panel containing the summary profile. "main" refers to the heatmap
        of individual values
    """
    fig, ax = get_fig_axes(axes)
    axes = split_axes(ax, top_height=0.2)

    if sort_fn is None:
        sort_indices = numpy.arange(data.shape[0])
    else:
        sort_indices = sort_fn(data)

    if x is None:
        x = numpy.arange(0, data.shape[1])

    if profile is None:
        profile = numpy.nanmedian(data, axis=0)

    im_args = copy.deepcopy(im_args)

    # populate with defaults
    for k, v in _heatmap_defaults.items():
        if k not in im_args:
            im_args[k] = v

    if "extent" not in im_args:
        im_args["extent"] = [x.min(), x.max(), 0, data.shape[0]]
    if "vmin" not in im_args:
        im_args["vmin"] = numpy.nanmin(data)
    if "vmax" not in im_args:
        im_args["vmax"] = numpy.nanmax(data)

    if cmap is not None:
        im_args["cmap"] = cmap
    elif "cmap" in im_args:
        cmap = matplotlib.cm.get_cmap(im_args["cmap"])
    else:
        cmap = matplotlib.cm.get_cmap()

    cmap.set_bad(nancolor, 1.0)

    axes["top"].plot(x, profile, **plot_args)
    axes["top"].set_ylim(0, profile.max())
    axes["top"].set_xlim(x.min(), x.max())
    #axes["top"].set_yticks([])
    axes["top"].set_yticks([0, profile.max()])
    axes["top"].xaxis.tick_bottom()
    axes["top"].grid(True, which="both")

    axes["main"].xaxis.tick_bottom()
    axes["main"].imshow(data[sort_indices, :], **im_args)

    return fig, axes


#==============================================================================
# Scatter plots with marginal distributions
#==============================================================================

def _scatterhist_help(
        axes          = None,
        top_height    = 0,
        left_width    = 0,
        right_width   = 0,
        bottom_height = 0,
): # yapf: disable
    """Create a scatter plot with the marginal distribution for `x`
    plotted in a separate pain as a kernel density estimate.

    Parameters
    ----------
    color : matplotlib colorspec, or None, optional
        Color to plot data series

    axes : :class:`matplotlib.axes.Axes`, dict, or `None`, optional
        If a :class:`matplotlib.axes.Axes`, an axes in which to place plot.
        This axes will be split into relevant panels.

        If a dict, this is expected to be a group of pre-split axes.

        If `None`, a new figure is generated, and axes are split.
        (Default: `None`)
      
    top_height, left_width, right_width, bottom_height : float, optional
        If not `None`, a panel on the corresponding side of the `ax` will
        be created, using whatever fraction is specified (e.g. 0.1 to use
        10% of total height).    


    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Parent figure of axes
    
    dict of :class:`matplotlib.axes.Axes`
        axes containing plot space
    """
    if axes is None:
        fig = plt.figure()
        axes = plt.gca()
    if isinstance(axes, matplotlib.axes.Axes):
        fig = axes.figure
        axes = split_axes(
            axes,
            top_height=top_height,
            left_width=left_width,
            right_width=right_width,
            bottom_height=bottom_height
        )
    elif isinstance(axes, dict):
        fig = axes["main"].figure
        if left_width > 0:
            assert "left" in axes
        if right_width > 0:
            assert "right" in axes
        if bottom_height > 0:
            assert "bottom" in axes
        if top_height > 0:
            assert "top" in axes

    return fig, axes

def scatterhist_x(
        x,
        y,
        color        = None,
        axes         = None,
        label        = None,
        top_height   = 0.2,
        mask_invalid = True,
        log          = "",
        min_x        = -numpy.inf,
        min_y        = -numpy.inf,
        max_x        = numpy.inf,
        max_y        = numpy.inf,
        scargs       = plastid_default_scatter,
        bw_method    = "scott",
        kdalpha      = 0.7
): # yapf: disable
    """Produce a scatter plot with a kernel density estimate of the marginal `x` distribution

    Parameters
    ----------
    x, y : :class:`numpy.ndarray` or list
        Pair arrays or lists of corresponding numbers

    color : matplotlib colorspec, optional
        Color to use in plot

    label : str, or None
        If not None, a label for plotting

    axes : :class:`matplotlib.axes.Axes`, dict, or `None`, optional
        If a :class:`matplotlib.axes.Axes`, an axes in which to place plot.
        This axes will be split into relevant panels.

        If a dict, this is expected to be a group of pre-split axes.

        If `None`, a new figure is generated, and axes are split.
        (Default: `None`)

    top_height : float, optional
        fraction of `axes` height to use in top panel containing
        marginal distribution (Default: 0.2)

    mask_invalid : bool, optional
        If `True` mask out any `nan`s or `inf`s, as these mess up kernel density
        estimates and histograms in matplotlib, even if in masked arrays

    min_x, min_y, max_x, max_y : number, optional
        If supplied, set values below `min_x` to `min_x`, values larger
        than `max_x` to `max_x` and so for `min_y` and `max_y`

    log : str, "", "x", "xy", or "xy", optional
        Plot these axes on a log scale (Default: "" for no log axes)

    scargs : Keyword arguments, optional
        Arguments to pass to :func:`~matplotlib.pyplot.scatter`
        (Default: :obj:`plastid_default_scatter`). We highly recommend
        setting `rasterized` to `True`!

    kdalpha : float, optional
        Alpha level (transparency) for marginal distributions (Default: 0.7)

    bw_method : str
        Bandwith estimation method. See documentation for
        :obj:`scipy.stats.gaussian_kde`. (Default: `"scott"`)


    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Figure

    dict
        Dictionary of axes. `'orig'` refers to `ax`. The central panel is `'main'`.
        Other panels will be mapped to `'top'`, `'left'` et c, if they are created.
    """
    fig, axes = _scatterhist_help(axes=axes, top_height=top_height)
    xlog = False

    if color is None:
        color = next(get_color_cycle(axes["main"]))

    if mask_invalid == True:
        x, y = clean_invalid(x, y, min_x=min_x, max_x=max_x, min_y=min_y, max_y=max_y)

    if label is not None:
        scargs = copy.deepcopy(scargs)
        scargs["label"] = label

    if "x" in log:
        axes["main"].semilogx()
        axes["top"].semilogx()
        xlog = True
        xmask = x > 0
    else:
        xmask = numpy.tile(True, x.shape)

    if "y" in log:
        axes["main"].semilogy()

    axes["main"].scatter(x, y, edgecolor=color, **scargs)

    # arguments for kernel desntiy plot
    kargs = {
        "color": color,
        "alpha": kdalpha,
        "bw_method": bw_method,
    }

    if xmask.sum() > 0:
        kde_plot(x[xmask], log=xlog, axes=axes["top"], **kargs)

    return fig, axes

def scatterhist_y(
        x,
        y,
        color        = None,
        axes         = None,
        label        = None,
        right_width  = 0.2,
        mask_invalid = True,
        log          = "xy",
        min_x        = -numpy.inf,
        min_y        = -numpy.inf,
        max_x        = numpy.inf,
        max_y        = numpy.inf,
        scargs       = plastid_default_scatter,
        bw_method    = "scott",
        kdalpha      = 0.7
): # yapf: disable
    """Produce a scatter plot with a kernel density estimate of the marginal `y` distribution

    Parameters
    ----------
    x, y : :class:`numpy.ndarray` or list
        Pair arrays or lists of corresponding numbers

    color : matplotlib colorspec, optional
        Color to use in plot

    label : str, or None
        If not None, a label for plotting

    axes : :class:`matplotlib.axes.Axes`, dict, or `None`, optional
        If a :class:`matplotlib.axes.Axes`, an axes in which to place plot.
        This axes will be split into relevant panels.

        If a dict, this is expected to be a group of pre-split axes.

        If `None`, a new figure is generated, and axes are split.
        (Default: `None`)

    right_width : float, optional
        fraction of `axes` width to use in right panel containing
        marginal distribution (Default: 0.2)

    mask_invalid : bool, optional
        If `True` mask out any `nan`s or `inf`s, as these mess up kernel density
        estimates and histograms in matplotlib, even if in masked arrays

    min_x, min_y, max_x, max_y : number, optional
        If supplied, set values below `min_x` to `min_x`, values larger
        than `max_x` to `max_x` and so for `min_y` and `max_y`

    log : str, "", "x", "xy", or "xy", optional
        Plot these axes on a log scale (Default: "" for no log axes)

    scargs : Keyword arguments, optional
        Arguments to pass to :func:`~matplotlib.pyplot.scatter`
        (Default: :obj:`plastid_default_scatter`). We highly recommend
        setting `rasterized` to `True`!

    kdalpha : float, optional
        Alpha level (transparency) for marginal distributions (Default: 0.7)

    bw_method : str
        Bandwith estimation method. See documentation for
        :obj:`scipy.stats.gaussian_kde`. (Default: "scott")

    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Figure

    dict
        Dictionary of axes. `'orig'` refers to `ax`. The central panel is `'main'`.
        Other panels will be mapped to `'top'`, `'left'` et c, if they are created.
    """
    fig, axes = _scatterhist_help(axes=axes, right_width=right_width)
    ylog = False

    if color is None:
        color = next(get_color_cycle(axes["main"]))

    if label is not None:
        scargs = copy.deepcopy(scargs)
        scargs["label"] = label

    if mask_invalid == True:
        x, y = clean_invalid(x, y, min_x=min_x, max_x=max_x, min_y=min_y, max_y=max_y)

    if "x" in log:
        axes["main"].semilogx()

    if "y" in log:
        axes["main"].semilogy()
        axes["right"].semilogy()
        ylog = True
        ymask = y > 0
    else:
        ymask = numpy.tile(True, y.shape)

    axes["main"].scatter(x, y, edgecolor=color, **scargs)

    # kernel density
    kargs = {
        "color": color,
        "alpha": kdalpha,
        "bw_method": bw_method,
    }

    if ymask.sum() > 0:
        kde_plot(y[ymask], log=ylog, axes=axes["right"], vert=True, **kargs)

    return fig, axes

def scatterhist_xy(
        x,
        y,
        color        = None,
        axes         = None,
        label        = None,
        top_height   = 0.2,
        right_width  = 0.2,
        mask_invalid = True,
        log          = "xy",
        min_x        = -numpy.inf,
        min_y        = -numpy.inf,
        max_x        = numpy.inf,
        max_y        = numpy.inf,
        scargs       = plastid_default_scatter,
        kdalpha      = 0.7,
        bw_method    = "scott"
): # yapf: disable
    """Produce a scatter plot with kernel density estimate of the marginal `x` and `y` distributions

    Parameters
    ----------
    x, y : :class:`numpy.ndarray` or list
        Pair arrays or lists of corresponding numbers

    color : matplotlib colorspec, optional
        Color to use in plot

    label : str, or None
        If not None, a label for plotting    top_height : float, optional
        fraction of `axes` height to use in top panel containing
        marginal distribution (Default: 0.2)
        
    right_width : float, optional
        fraction of `axes` width to use in right panel containing
        marginal distribution (Default: 0.2)

    top_height : float, optional
        fraction of `axes` height to use in top panel containing
        marginal distribution (Default: 0.2)

    axes : :class:`matplotlib.axes.Axes`, dict, or `None`, optional
        If a :class:`matplotlib.axes.Axes`, an axes in which to place plot.
        This axes will be split into relevant panels.

        If a dict, this is expected to be a group of pre-split axes.

        If `None`, a new figure is generated, and axes are split.
        (Default: `None`)

    mask_invalid : bool, optional
        If `True` mask out any `nan`s or `inf`s, as these mess up kernel density
        estimates and histograms in matplotlib, even if in masked arrays

    min_x, min_y, max_x, max_y : number, optional
        If supplied, set values below `min_x` to `min_x`, values larger
        than `max_x` to `max_x` and so for `min_y` and `max-y`

    log : str, "", "x", "xy", or "xy", optional
        Plot these axes on a log scale (Default: "" for no log axes)

    scargs : Keyword arguments, optional
        Arguments to pass to :func:`~matplotlib.pyplot.scatter`
        (Default: :obj:`plastid_default_scatter`). We highly recommend
        setting `rasterized` to `True`!

    kdalpha : float, optional
        Alpha level (transparency) for marginal distributions (Default: 0.7)

    bw_method : str
        Bandwith estimation method. See documentation for
        :obj:`scipy.stats.gaussian_kde`. (Default: "scott")

    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Figure

    dict
        Dictionary of axes. `'orig'` refers to `ax`. The central panel is `'main'`.
        Other panels will be mapped to `'top'`, `'left`' et c, if they are created.
    """
    fig, axes = _scatterhist_help(axes=axes, top_height=top_height, right_width=right_width)
    xlog = ylog = False

    if color is None:
        color = next(get_color_cycle(axes["main"]))

    if mask_invalid == True:
        x, y = clean_invalid(x, y, min_x=min_x, max_x=max_x, min_y=min_y, max_y=max_y)

    if label is not None:
        scargs = copy.deepcopy(scargs)
        scargs["label"] = label

    if "x" in log:
        axes["main"].semilogx()
        axes["top"].semilogx()
        xlog = True
        xmask = x > 0
    else:
        xmask = numpy.tile(True, x.shape)

    if "y" in log:
        axes["main"].semilogy()
        axes["right"].semilogy()
        ylog = True
        ymask = y > 0
    else:
        ymask = numpy.tile(True, y.shape)

    axes["main"].scatter(x, y, edgecolor=color, **scargs)

    # kernel densities
    kargs = {
        "color": color,
        "alpha": kdalpha,
        "bw_method": bw_method,
    }

    if ymask.sum() > 0:
        kde_plot(y[ymask], log=ylog, axes=axes["right"], vert=True, **kargs)

    if xmask.sum() > 0:
        kde_plot(x[xmask], log=xlog, axes=axes["top"], **kargs)

    return fig, axes


#==============================================================================
# Plots specific for genomics
#==============================================================================

def ma_plot(
        x,
        y,
        axes         = None,
        color        = None,
        label        = None,
        xlabel       = None,
        ylabel       = None,
        title        = None,
        right_width  = 0.2,
        log          = "xy",
        min_x        = -numpy.inf,
        max_x        = numpy.inf,
        min_y        = -numpy.inf,
        max_y        = numpy.inf,
        scargs       = plastid_default_scatter,
        mask_invalid = True,
        kdalpha      = 0.7
): # yapf: disable
    """Plot fold changes (:math:`\log_{2} (y/x)`) as a function of the mean of x and y (:math:`0.5*(x+y)`).

    Parameters
    ----------
    x, y : :class:`numpy.ndarray` or list
        Pair arrays or lists of corresponding numbers

    color : matplotlib colorspec, optional
        Color to use in plot

    label : str or None, optional
        If not `None`, a label for plotting

    xlabel : str or None, optional
        If not `None`, an x-axis label

    ylabel : str or None, optional
        If not `None`, a y-axis label
        
    right_width : float, optional
        fraction of `axes` width to use in right panel containing
        marginal distribution (Default: 0.2)

    axes : :class:`matplotlib.axes.Axes`, dict, or `None`, optional
        If a :class:`matplotlib.axes.Axes`, an axes in which to place plot.
        This axes will be split into relevant panels.

        If a dict, this is expected to be a group of pre-split axes.

        If `None`, a new figure is generated, and axes are split.
        (Default: `None`)

    mask_invalid : bool, optional
        If `True` mask out any `nan`s or `inf`s, as these mess up kernel density
        estimates and histograms in matplotlib, even if in masked arrays

    min_x, min_y, max_x, max_y : number, optional
        If supplied, set values below `min_x` to `min_x`, values larger
        than `max_x` to `max_x` and so for `min_y` and `max_y`

    log : str, "", "x", "xy", or "xy", optional
        Plot these axes on a log scale (Default: "xy")

    scargs : Keyword arguments, optional
        Arguments to pass to :func:`~matplotlib.pyplot.scatter`
        (Default: :obj:`plastid_default_scatter` ). Recommend: set `rasterized`
        to `True`

    kdalpha : float, optional
        Alpha level (transparency) for marginal distributions (Default: 0.7)


    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Figure

    dict
        Dictionary of axes. `'orig'` refers to `ax`. The central panel is `'main'`.
        Other panels will be mapped to `'top'`, `'left'` et c, if they are created.
    """

    do_setup = axes is None
    logs = numpy.ma.masked_invalid(numpy.log2(y / x))
    imask = ~logs.mask

    ratio = y / x
    mean = 0.5 * (x + y)

    fig, axdict = scatterhist_y(
        mean[imask],
        ratio[imask],
        axes         = axes,
        min_x        = min_x,
        max_x        = max_x,
        min_y        = min_y,
        max_y        = max_y,
        log          = log,
        right_width  = right_width,
        color        = color,
        mask_invalid = mask_invalid,
        label        = label,
        kdalpha      = kdalpha
    ) # yapf: disable

    if do_setup == True:
        axdict["main"].axhline(1, color=process_black, zorder=-5, linewidth=1)
        axdict["right"].axhline(1, color=process_black, zorder=-5, linewidth=1)
        axdict["right"].xaxis.set_ticklabels([])

        if ylabel is None:
            ylabel = "log2 fold change"

        axdict["main"].set_ylabel(ylabel)
        if xlabel is not None:
            axdict["main"].set_xlabel(xlabel)

        if title is not None:
            plt.suptitle(title)

    return fig, axdict


def phase_plot(counts, labels=None, cmap=None, color=None, lighten_by=0.2, fig={}, line={}, bar={}):
    """Phasing plot for ribosome profiling

    Creates a two-panel plot:

      - the top panel is a line graph indicating the fraction of reads
        as a function of read length

      - the bottom panel is a stacked bar graph, showing the fraction
        of reads in each codon position for each read length, with
        codon position 2 stacked above position 1 stacked above position 0


    Parameters
    ----------
    counts : :class:`numpy.ndarray`
        Nx3 array of raw counts, where each row represents a read length,
        and each column a codon phase

    labels : list, optional
        Labels for each stack. If `None`, stacks are labeled sequentially by number.
        (Default: `None`)
    
    lighten_by : float, optional
        Amount by which to lighten sequential blocks in each stack. (Default: 0.10)
    
    cmap : :class:`matplotlib.colors.Colormap`, optional
        Colormap from which to generate bar colors. If supplied, will override
        any `color` attribute in `**kwargs`.

    color : matplotlib colorspec, or list of these
        Colors to use in plot. Overridden if `cmap` is supplied.

    fig : dict
        Keyword arguments to :func:`matplotlib.pylot.figure`
        
    line : dict
        Keyword arguments to pass to :func:`matplotlib.pyplot.plot`
        in top panel

    bar : dict
        Keyword arguments to pass to :func:`matplotlib.pyplot.bar`
        in bottom panel


    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Figure

    tuple
        Tuple of :class:`matplotlib.axes.Axes`; the first corresponding 
        to the line graph (top panel), the second, the bar graph (bottom).
    """
    fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, sharex=True, **fig)

    totals = counts.sum(1)
    phases = (counts.astype(float).T / totals).T

    stacked_bar(
        phases, axes=ax2, labels=labels, lighten_by=lighten_by, cmap=cmap, color=color, **bar
    )
    ax2.set_xlabel("Read length (nt)")
    ax2.set_ylabel("Fraction in each phase")
    x = numpy.arange(len(totals)) + 0.5

    if "color" not in line:
        line["color"] = process_black

    ax1.plot(x, (totals.astype(float)) / totals.sum(), **line)

    ax1.set_ylabel("Fraction of reads")

    return fig, (ax1, ax2)



#!/usr/bin/env python
"""This module defines a two classes for reading `PSL`_ files (made by, for example,
`blat`_):


|PSL_Reader|
    Read a `PSL`_ file line-by-line, converting each line into a |SegmentChain|
    or |Transcript|

|BundledPSL_Reader|
    Read `PSL`_ files, returning lists of |SegmentChains| grouped by query sequence.
"""
__date__ = "2011-09-01"
__author__ = "joshua"
import itertools
from plastid.readers.common import AssembledFeatureReader
from plastid.genomics.roitools import SegmentChain
from plastid.util.services.exceptions import FileFormatWarning, warn


class PSL_Reader(AssembledFeatureReader):
    """
    PSL_reader(*streams, return_type=SegmentChain, add_three_for_stop=False, tabix=False, printer=None, **kwargs)
    
    Read `PSL`_ files into |SegmentChain|  or |Transcript| objects


    Parameters
    ----------
    *streams : file-like
        One or more open filehandles of input data.
    
    return_type : |SegmentChain| or subclass, optional
        Type of feature to return from assembled subfeatures (Default: |SegmentChain|)

    add_three_for_stop : bool, optional
        Some annotation files exclude the stop codon from CDS annotations. If set to
        `True`, three nucleotides will be added to the threeprime end of each
        CDS annotation, **UNLESS** the annotated transcript contains explicit stop_codon 
        feature. (Default: `False`)
                    
    printer : file-like, optional
        Logger implementing a ``write()`` method. Default: |NullWriter|

    tabix : bool, optional
        `streams` point to `tabix`_-compressed files or are open
        :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)


    **kwargs
        Other keyword arguments used by specific parsers

    
    Attributes
    ----------
    streams : file-like
        One or more open streams (usually filehandles) of input data.
    
    return_type : class
        The type of object assembled by the reader. Typically an |SegmentChain|
        or a subclass thereof. Must import a method called ``from_psl``

    counter : int
        Cumulative line number counter over all streams
    
    rejected : list
        A list of lines from `PSL`_ file that did not assemble properly
    
    metadata : dict
        Various attributes gleaned from the stream, if any    
    """

    def _assemble(self, line):
        """Read `PSL`_ files line-by-line into types specified by ``self.return_type``"""
        self.counter += 1
        if line.strip() == "":
            return self.__next__()
        elif line.startswith("psLayout"):
            return self.__next__()
        elif line.lstrip().startswith("match"):
            return self.__next__()
        elif line.startswith("--"):
            return self.__next__()
        elif line.startswith("#"):
            return self.__next__()
        else:
            try:
                return self.return_type.from_psl(line)
            except Exception as e:
                self.rejected.append(line)
                warn(
                    "Rejecting line %s because of %s: %s" % (self.counter, e.message, line),
                    FileFormatWarning
                )
                return self.__next__()


class BundledPSL_Reader(PSL_Reader):
    """
    BundledPSL_reader(*streams, return_type=SegmentChain, add_three_for_stop=False, tabix=False, printer=None, **kwargs)
    
    Read `PSL`_ files, returning lists of |SegmentChains| grouped by query sequence.
    Use this when a given query sequence has multiple hits in your `PSL`_ file,
    and you want the output to be grouped.
    

    Parameters
    ----------
    *streams : file-like
        One or more open filehandles of input data.
    
    return_type : |SegmentChain| or subclass, optional
        Type of feature to return from assembled subfeatures (Default: |SegmentChain|)

    add_three_for_stop : bool, optional
        Some annotation files exclude the stop codon from CDS annotations. If set to
        `True`, three nucleotides will be added to the threeprime end of each
        CDS annotation, **UNLESS** the annotated transcript contains explicit stop_codon 
        feature. (Default: `False`)
                    
    printer : file-like, optional
        Logger implementing a ``write()`` method. Default: |NullWriter|
    
    tabix : bool, optional
        `streams` point to `tabix`_-compressed files or are open
        :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)

    **kwargs
        Other keyword arguments used by specific parsers    
    """

    def filter(self, line):
        """Process lines of `PSL`_ files input into |SegmentChain|, and group
        these by query sequence.
         
        Parameters
        ----------
        line : str
            line of `PSL`_ input
         
        Returns
        -------
        list
            list of |SegmentChain| objects sharing a query sequence 
        """
        ltmp = []
        aln = SegmentChain.from_psl(line)
        last_name = aln.attr["query_name"]
        try:
            while last_name == aln.attr["query_name"]:
                ltmp.append(aln)
                line = next(self.stream)
                aln = SegmentChain.from_psl(line)

            self.stream = itertools.chain([line], self.stream)
            return ltmp
        except StopIteration:
            # send final bundle
            return ltmp



#!/usr/bin/env python
"""Tests for data structures defined in :py:mod:`plastid.readers.common`
"""
import unittest
import pysam
import copy
from nose.plugins.attrib import attr
from plastid.readers.common import add_three_for_stop_codon, \
                                   get_identical_attributes, \
                                   AssembledFeatureReader
from plastid.test.ref_files import REF_FILES
from plastid.genomics.roitools import GenomicSegment,\
                                      SegmentChain, \
                                      Transcript


@attr(test="unit")
class TestGetIdenticalAttributes(unittest.TestCase):
    """Tests :py:func:`get_identical_attributes`"""

    @classmethod
    def setUpClass(cls):
        cls.ivs = [
            GenomicSegment("chrA", 100, 190, "+"),
            GenomicSegment("chrA", 200, 203, "+"),
            GenomicSegment("chrA", 200, 201, "+"),
            GenomicSegment("chrA", 204, 206, "+"),
        ]
        cls.common_attr = dict(common1="common", common2="also common", common3="still common")
        cls.attrs = [
            dict(common_diff_val="unique_f1", unique_f1_key="something"),
            dict(
                common_diff_val="unique_f2",
                unique_f2_key="something",
                unique_f2f3="something else"
            ),
            dict(
                common_diff_val="unique_f3",
                unique_f3_key="something",
                unique_f2f3="something else",
                unique_f3f4="f3 only"
            ),
            dict(common_diff_val="unique_f4", unique_f4_key="something", unique_f3f4="f4 only"),
        ]
        for x in cls.attrs:
            x.update(cls.common_attr)

    def test_on_SegmentChain_no_exclude(self):
        features = [SegmentChain(self.ivs[n], **self.attrs[n]) for n in range(len(self.ivs))]
        common_plus_type = copy.deepcopy(self.common_attr)
        common_plus_type["type"] = "exon"
        self.assertEqual(get_identical_attributes(features), common_plus_type)

    def test_on_SegmentChain_exclude(self):
        features = [SegmentChain(self.ivs[n], **self.attrs[n]) for n in range(len(self.ivs))]
        self.assertEqual(get_identical_attributes(features, exclude=["type"]), self.common_attr)


class MockReader(AssembledFeatureReader):
    """AssembledFeatureReader that returns string literals from files"""

    def _assemble(self, data):
        return data


class test_AssembledFeatureReader(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tabix_ref = REF_FILES["100transcripts_tabixbed"]
        cls.bed_ref = REF_FILES["100transcripts_bed"]
        with open(cls.bed_ref) as bed_fh:
            cls.bed_lines = bed_fh.readlines()

    # methods below test tabix keyword on opening strings, open filehandles,
    # or pysam.tabix_iterators by comparing string output to expected string literals
    def test_tabix_str_open(self):
        reader = MockReader(self.tabix_ref, tabix=True)
        for expected, found in zip(reader, self.bed_lines):
            self.assertEqual(expected.strip("\n"), found.strip("\n"))

    def test_tabix_multi_str_open(self):
        reader = MockReader(self.tabix_ref, self.tabix_ref, tabix=True)
        for expected, found in zip(reader, self.bed_lines + self.bed_lines):
            self.assertEqual(expected.strip("\n"), found.strip("\n"))

    def test_tabix_fh_open(self):
        with open(self.tabix_ref, "rb") as fh1:
            reader = MockReader(fh1, self.tabix_ref, tabix=True)
            for expected, found in zip(reader, self.bed_lines):
                self.assertEqual(expected.strip("\n"), found.strip("\n"))

    def test_tabix_multi_fh_open(self):
        with open(self.tabix_ref, "rb") as fh1:
            with open(self.tabix_ref, "rb") as fh2:
                reader = MockReader(fh1, fh2, self.tabix_ref, tabix=True)
                for expected, found in zip(reader, self.bed_lines + self.bed_lines):
                    self.assertEqual(expected.strip("\n"), found.strip("\n"))

    def test_tabix_ps_open(self):
        with open(self.tabix_ref, "rb") as fh1:
            ps1 = pysam.tabix_file_iterator(fh1, pysam.asTuple())
            reader = MockReader(ps1, self.tabix_ref, tabix=True)
            for expected, found in zip(reader, self.bed_lines):
                self.assertEqual(expected.strip("\n"), found.strip("\n"))

    def test_tabix_multi_ps_open(self):
        with open(self.tabix_ref, "rb") as fh1:
            with open(self.tabix_ref, "rb") as fh2:
                ps1 = pysam.tabix_file_iterator(fh1, pysam.asTuple())
                ps2 = pysam.tabix_file_iterator(fh2, pysam.asTuple())
                reader = MockReader(ps1, ps2, self.tabix_ref, tabix=True)
                for expected, found in zip(reader, self.bed_lines + self.bed_lines):
                    self.assertEqual(expected.strip("\n"), found.strip("\n"))



#!/usr/bin/env python
"""Test suite for filters in :py:mod:`plastid.util.io.filters`
"""
import unittest
from nose.plugins.attrib import attr
from plastid.util.services.mini2to3 import cStringIO
from plastid.util.io.filters import SkipBlankReader, CommentReader, FunctionReader

#===============================================================================
# INDEX: test cases
#===============================================================================


@attr(test="unit")
class TestSkipBlankReader(unittest.TestCase):
    """TestCase for :py:class:`SkipBlankReader`"""

    def get_stream(self):
        return SkipBlankReader(cStringIO.StringIO(_NL_TEXT))

    def test_read(self):
        fin = self.get_stream()
        lines = fin.read().strip().split("\n")
        self.assertEqual(len(lines), 30)
        for line in lines:
            self.assertNotEqual(line.strip(), "")
        fin.close()

    def test_readlines(self):
        lines = self.get_stream().readlines()
        self.assertEqual(len(lines), 30)
        for line in lines:
            self.assertNotEqual(line.strip(), "")

    def test_iter(self):
        for n, line in enumerate(self.get_stream()):
            self.assertNotEqual(line.strip(), "")

        self.assertEqual(n, 29)


@attr(test="unit")
class TestCommentReader(TestSkipBlankReader):
    """TestCase for :py:class:`CommentReader`"""

    def get_stream(self):
        return CommentReader(cStringIO.StringIO(_COMMENT_TEXT))

    def test_get_comments(self):
        reader = self.get_stream()
        comments = reader.get_comments()
        # comments should not be populated until we have read
        self.assertEquals(len(comments), 0)
        reader.read()
        # comments should be found afterwards
        self.assertEquals(len(comments), 5)
        for comment in comments:
            self.assertTrue(comment.startswith("#"))


@attr(test="unit")
class TestFunctionReader(unittest.TestCase):
    def test_via_backwards(self):
        reader = FunctionReader(cStringIO.StringIO(_NL_TEXT), lambda x: x[::-1])
        for line1, line2 in zip(reader, cStringIO.StringIO(_NL_TEXT)):
            self.assertEqual(line1, line2[::-1])

    def test_via_upper(self):
        reader = FunctionReader(cStringIO.StringIO(_NL_TEXT), str.upper)
        for line1, line2 in zip(reader, cStringIO.StringIO(_NL_TEXT)):
            self.assertEqual(line1, line2.upper())


#===============================================================================
# INDEX: test data
#===============================================================================

_NL_TEXT = """1	Lorem ipsum dolor sit amet, consectetur adipiscing elit. Morbi vitae ipsum vel
     2	nisi dapibus dapibus in vel neque. Proin nec consectetur arcu. Praesent
     3	convallis venenatis metus quis elementum. Ut elementum justo ac faucibus
     4	efficitur. Nam at ornare elit. Sed pulvinar mi sapien, sed faucibus risus
     5	cursus at. Nulla sit amet posuere ex, sit amet convallis ante. Integer
     6	imperdiet purus nec ante pretium fringilla. Vivamus a ligula tristique,
     7	sodales elit et, fringilla ligula. Nulla facilisi. Nunc sed enim non ligula
     8	commodo blandit. Sed ultricies quis urna vel imperdiet. Curabitur leo nisi,
     9	faucibus sed mauris ut, bibendum pretium nibh.
       
    10	Cras libero quam, scelerisque ut suscipit ac, pellentesque non sapien. Vivamus
    11	eu tristique nisi. Sed luctus molestie mollis. Praesent dapibus tincidunt
    12	pretium. Sed faucibus vestibulum est, ac mollis libero dapibus in. Phasellus
    13	nec euismod sapien. Donec faucibus orci sem, vitae hendrerit orci sodales non.
    14	Sed facilisis erat at erat semper facilisis. Proin at orci et ligula mattis
    15	condimentum. Aenean tincidunt, nunc sit amet malesuada scelerisque, nisl augue
    16	imperdiet lectus, et sollicitudin quam quam nec massa. Morbi volutpat nulla et
    17	erat porta tempus.
       
    18	Nulla placerat ipsum elit, at vestibulum urna pellentesque a. Nunc bibendum
    19	convallis orci, vel euismod tellus commodo ullamcorper. Vestibulum vestibulum
    20	lobortis tempor. Maecenas non nisi aliquet lorem tincidunt varius sed aliquam
    21	est. Vivamus egestas, nisi sed laoreet interdum, nisi tellus egestas erat, sit
    22	amet ultricies metus orci sed orci. Aenean at lectus viverra, venenatis mauris
    23	sed, ultricies augue. Sed ac purus vitae dui tempus pulvinar. Duis tincidunt
                               
    24	nisi sit amet purus laoreet placerat. Quisque nulla orci, vestibulum nec metus
    25	eu, convallis consequat libero. Maecenas sed ultrices tellus, non commodo
    26	lectus. Suspendisse non diam eget purus imperdiet aliquam a sed libero.
    									
    27	Quisque eu venenatis tellus. Aliquam ornare purus non ante imperdiet interdum.
    28	Cras pretium, diam ac sagittis luctus, diam metus scelerisque libero, sit amet
                           29	cursus erat nulla ut augue. Ut est tortor, faucibus non nisi a, gravida
    30	fermentum tortor.



"""

_COMMENT_TEXT = """1	Lorem ipsum dolor sit amet, consectetur adipiscing elit. Morbi vitae ipsum vel
     2	nisi dapibus dapibus in vel neque. Proin nec consectetur arcu. Praesent
     3	convallis venenatis metus quis elementum. Ut elementum justo ac faucibus
     4	efficitur. Nam at ornare elit. Sed pulvinar mi sapien, sed faucibus risus
     5	cursus at. Nulla sit amet posuere ex, sit amet convallis ante. Integer
     6	imperdiet purus nec ante pretium fringilla. Vivamus a ligula tristique,
     7	sodales elit et, fringilla ligula. Nulla facilisi. Nunc sed enim non ligula
     8	commodo blandit. Sed ultricies quis urna vel imperdiet. Curabitur leo nisi,
     9	faucibus sed mauris ut, bibendum pretium nibh.
# Here is comment #1  
    10	Cras libero quam, scelerisque ut suscipit ac, pellentesque non sapien. Vivamus
    11	eu tristique nisi. Sed luctus molestie mollis. Praesent dapibus tincidunt
    12	pretium. Sed faucibus vestibulum est, ac mollis libero dapibus in. Phasellus
    13	nec euismod sapien. Donec faucibus orci sem, vitae hendrerit orci sodales non.
    14	Sed facilisis erat at erat semper facilisis. Proin at orci et ligula mattis
    15	condimentum. Aenean tincidunt, nunc sit amet malesuada scelerisque, nisl augue
    16	imperdiet lectus, et sollicitudin quam quam nec massa. Morbi volutpat nulla et
    17	erat porta tempus.
# This is comment #2
    18	Nulla placerat ipsum elit, at vestibulum urna pellentesque a. Nunc bibendum
    19	convallis orci, vel euismod tellus commodo ullamcorper. Vestibulum vestibulum
    20	lobortis tempor. Maecenas non nisi aliquet lorem tincidunt varius sed aliquam
    21	est. Vivamus egestas, nisi sed laoreet interdum, nisi tellus egestas erat, sit
    22	amet ultricies metus orci sed orci. Aenean at lectus viverra, venenatis mauris
    23	sed, ultricies augue. Sed ac purus vitae dui tempus pulvinar. Duis tincidunt
# This is comment #3                               
    24	nisi sit amet purus laoreet placerat. Quisque nulla orci, vestibulum nec metus
    25	eu, convallis consequat libero. Maecenas sed ultrices tellus, non commodo  # this is an end-of-line comment, which should not be removed
    26	lectus. Suspendisse non diam eget purus imperdiet aliquam a sed libero.
# This is comment #4						
    27	Quisque eu venenatis tellus. Aliquam ornare purus non ante imperdiet interdum.
    28	Cras pretium, diam ac sagittis luctus, diam metus scelerisque libero, sit amet
                           29	cursus erat nulla ut augue. Ut est tortor, faucibus non nisi a, gravida
    30	fermentum tortor.
# This is comment #5"""



#!/usr/bin/env python
import unittest
import warnings
import copy
from nose.tools import assert_equal, assert_not_equal
from nose.plugins.attrib import attr
from plastid.readers.gff_tokens import (
    escape_GFF3,
    unescape_GFF3,
    escape_GTF2,
    unescape_GTF2,
    make_GFF3_tokens,
    make_GTF2_tokens,
    parse_GFF3_tokens,
    parse_GTF2_tokens,
)
from plastid.util.services.decorators import catch_warnings
import plastid.util.services.exceptions

#===============================================================================
# INDEX: test data
#===============================================================================

# string requiring escaping
lipsum = """Lorem %;%ipsum dolor "sit amet, consectetur adipiscing elit, sed do
eiusmod temporincididunt ut labore et dolore magna aliqua. Ut enim ad minim
veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo
!*&)$#@(*RVLKJE@*(!U_)$(@*!)JFKJSF)@!(#_!@!(*$9087325gasdfjk@&!*Y*(_::{|}@{!';@
consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum;
dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident,
sunt in culpa qui officia deserunt mollit anim id est laborum.@(#*@FDSFJLK@!="""

lipsum_gff_escaped = """Lorem %25%3B%25ipsum dolor "sit amet%2C consectetur adipiscing elit%2C sed do
eiusmod temporincididunt ut labore et dolore magna aliqua. Ut enim ad minim
veniam%2C quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo
!*%26)$#@(*RVLKJE@*(!U_)$(@*!)JFKJSF)@!(#_!@!(*$9087325gasdfjk@%26!*Y*(_::{|}@{!'%3B@
consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum%3B
dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident%2C
sunt in culpa qui officia deserunt mollit anim id est laborum.@(#*@FDSFJLK@!%3D""".replace(
    "\n", "%0A"
)

lipsum_gtf_escaped = """Lorem %25%3B%25ipsum dolor %22sit amet%2C consectetur adipiscing elit%2C sed do
eiusmod temporincididunt ut labore et dolore magna aliqua. Ut enim ad minim
veniam%2C quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo
!*%26)$#@(*RVLKJE@*(!U_)$(@*!)JFKJSF)@!(#_!@!(*$9087325gasdfjk@%26!*Y*(_::{|}@{!'%3B@
consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum%3B
dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident%2C
sunt in culpa qui officia deserunt mollit anim id est laborum.@(#*@FDSFJLK@!%3D""".replace(
    "\n", "%0A"
)

# string that escaping should not change
medstring = "A medium sized string with spaces et c"
medstring_escaped = "A medium sized string with spaces et c"

# tokens formatted in GFF3

_GFF3_REPEATED = [
    ("repeated token no trailing semicolon", 'a=1;c=3;b=2;e=5;d=4;z=26;c=7'),
    ("repeated token trailing semicolon", 'a=1;c=3;b=2;e=5;d=4;z=26;c=7;'),
]

_GTF2_REPEATED = [
    ("repeated token no trailing semicolon", 'a "1"; c "3"; b "2"; e "5"; d "4"; z "26"; c "7"'),
    ("repeated token trailing semicolon", 'a "1"; c "3"; b "2"; e "5"; d "4"; z "26"; c "7";'),
]

_GFF3_TOKENS = [
    ("simple values no trailing semicolon", 'a=1;c=3;b=2;e=5;d=4;z=26'),
    ("too much spacing no trailing semicolon", 'a=1;   c=3;b=2  ;e=5;d=4;z=26'),
    (
        "escaped strings no trailing semicolon",
        'a=5;note1=%s;note2=%s;note3=%s' % (medstring, medstring_escaped, lipsum_gff_escaped)
    ),
    ("simple values with trailing semicolon", 'a=1;c=3;b=2;e=5;d=4;z=26;'),
    ("too much spacing with trailing semicolon", 'a=1;   c=3;b=2  ;e=5;d=4;z=26;'),
    ("two value non-expanded", 'a=1;c=3,7;b=2;e=5;d=4;z=26;'),
    (
        "escaped strings with trailing semicolon",
        'a=5;note1=%s;note2=%s;note3=%s;' % (medstring, medstring_escaped, lipsum_gff_escaped)
    ),
    (
        "escaped key", 'c%%2Cd=5;a=5;note1=%s;note2=%s;note3=%s;' %
        (medstring, medstring_escaped, lipsum_gff_escaped)
    ),

    # keep GFF3-specific elements at the end, so they are not tripped in GTF2
    (
        "two value GFF3 expanded no trailing semicolon",
        'Parent=p1,p2;a=1;c=3,7;b=2;e=5;d=4;z=26;Alias=alias1,alias2'
    ),
    (
        "two value GFF3 expanded with trailing semicolon",
        'Parent=p1,p2;a=1;c=3,7;b=2;e=5;d=4;z=26;Alias=alias1,alias2;'
    ),
]

# tokens formatted in GTF2
_GTF2_TOKENS = [
    ("simple values no trailing semicolon", 'a "1"; c "3"; b "2"; e "5"; d "4"; z "26"'),
    (
        "too much spacing no trailing semicolon",
        'a     "1"; c    "3"; b "2"; e   "5"; d "4"; z "26"'
    ),
    (
        "escaped strings no trailing semicolon", 'a "5"; note1 "%s"; note2 "%s"; note3 "%s"' %
        (medstring, medstring_escaped, lipsum_gtf_escaped)
    ),
    ("simple values with trailing semicolon", 'a "1"; c "3"; b "2"; e "5"; d "4"; z "26;"'),
    (
        "too much spacing with trailing semicolon",
        'a     "1"; c    "3"; b "2"; e   "5"; d "4"; z "26";'
    ),
    ("two value non-expanded", 'a "1"; c "3,7"; b "2"; e "5"; d "4"; z "26";'),
    (
        "escaped strings with trailing semicolon", 'a "5"; note1 "%s"; note2 "%s"; note3 "%s";' %
        (medstring, medstring_escaped, lipsum_gtf_escaped)
    ),
    (
        "escaped key", 'c%%2Cd "5"; a "5"; note1 "%s"; note2 "%s"; note3 "%s";' %
        (medstring, medstring_escaped, lipsum_gtf_escaped)
    ),  # key must be escaped
]

_TOKEN_PARSE_REPEATED = [
    dict(a="1", c=["3", "7"], b="2", e="5", d='4', z="26"),  # repeated tokens, must be catenated
    dict(a="1", c=["3", "7"], b="2", e="5", d='4', z="26"),  # repeated tokens, must be catenated
]

_TOKEN_WRITE_REPEATED = [
    dict(a="1", c="3,7", b="2", e="5", d='4', z="26"),  # repeated tokens, must be catenated
    dict(a="1", c="3,7", b="2", e="5", d='4', z="26"),  # repeated tokens, must be catenated
]

# dicitonaries corresponding to token strings above
_GFF3_TOKEN_DICTS = [
    dict(a="1", c="3", b="2", e="5", d='4', z="26"),  # just values, no trailing semicolon
    dict(a="1", c="3", b="2", e="5", d='4', z="26"),  # too much spacing
    dict(
        a="5",
        note1=medstring,  # requires escaping
        note2=medstring,
        note3=lipsum
    ),
    dict(a="1", c="3", b="2", e="5", d='4', z="26"),  # just values, trainling semicolon
    dict(a="1", c="3", b="2", e="5", d='4', z="26"),  # too much spacing
    dict(a="1", c="3,7", b="2", e="5", d='4', z="26"),  # two value non-expanded
    dict(
        a="5",
        note1=medstring,  # requires escaping
        note2=medstring,
        note3=lipsum
    ),
    {
        "c,d": "5",
        "a": "5",
        "note1": medstring,
        "note2": medstring,
        "note3": lipsum
    },
    dict(
        Parent=["p1", "p2"],
        Alias=["alias1", "alias2"],
        a="1",
        c="3,7",
        b="2",
        e="5",
        d='4',
        z="26"
    ),
    dict(
        Parent=["p1", "p2"],
        Alias=["alias1", "alias2"],
        a="1",
        c="3,7",
        b="2",
        e="5",
        d='4',
        z="26"
    ),
]

_GTF2_TOKEN_DICTS = _GFF3_TOKEN_DICTS[:-2]

#===============================================================================
# INDEX: tests for token parsing & writing
#===============================================================================


@attr(test="unit")
def test_escape_GFF3():
    yield assert_equal, escape_GFF3(lipsum), lipsum_gff_escaped
    yield assert_equal, escape_GFF3(medstring), medstring_escaped
    yield assert_not_equal, lipsum, lipsum_gff_escaped


@attr(test="unit")
def test_unescape_GFF3():
    yield assert_equal, unescape_GFF3(lipsum_gff_escaped), lipsum
    yield assert_equal, unescape_GFF3(medstring_escaped), medstring


@attr(test="unit")
def test_escape_GTF2():
    yield assert_equal, escape_GTF2(lipsum), lipsum_gtf_escaped
    yield assert_equal, escape_GTF2(medstring), medstring_escaped
    yield assert_not_equal, lipsum, lipsum_gtf_escaped


@attr(test="unit")
def test_unescape_GTF2():
    yield assert_equal, unescape_GTF2(lipsum_gtf_escaped), lipsum
    yield assert_equal, unescape_GTF2(medstring_escaped), medstring


@attr(test="unit")
class TestGFF3_TokenParsing(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.fmt = "GFF3"
        cls.tokens = _GFF3_TOKENS
        cls.token_dicts = _GFF3_TOKEN_DICTS
        cls.repeated = _GFF3_REPEATED
        cls.tokenizer = staticmethod(make_GFF3_tokens)
        cls.parser = staticmethod(parse_GFF3_tokens)

    def test_token_parse(self):
        pass
        """Test parsing of tokens in *attributes* column of GFF"""
        # test that all elements are in dict
        for c, (my_dict, (my_title, my_tokens)) in enumerate(zip(self.token_dicts, self.tokens)):
            found = self.parser(my_tokens)
            msg = "Failed %s parsing test '%s'\n    Expected: %s\n    Found: %s" % (
                self.fmt, my_title, my_dict, found
            )
            self.assertEqual(my_dict, found, msg)

        self.assertEqual(
            c,
            len(self.tokens) - 1,
            "Not all tokens parsed during test. Expected %s. Parsed %s" % (len(self.tokens) - 1, c)
        )

    def test_token_write_all(self, dicts=None):
        pass
        """Test export of GFF3 tokens
         
        Parameters
        ----------
        dicts : list<dict>
            List of dictionaries to convert to GFF3 tokens
        """
        if dicts is None:
            dicts = self.token_dicts

        # test total export from written tokens
        for c, my_dict in enumerate(dicts):
            my_str = self.tokenizer(my_dict, escape=True)
            new_dict = self.parser(my_str)
            self.assertEqual(my_dict, new_dict)

        self.assertGreater(c, 0)

    def test_token_write_subset(self, dicts=None, non_excludes=[]):
        pass
        """Test export of GFF3 tokens, excluding random subsets
        
        Parameters
        ----------
        dicts : list<dict>
            List of dictionaries to convert to GFF3 tokens
        
        non_excludes : list<str>
            Keys that cannot be excluded from export
        """
        # test excluding random key subsets from written tokens
        if dicts is None:
            dicts = self.token_dicts

        for c, old_dict in enumerate(dicts):

            excludes = set(list(old_dict.keys())[0:3]) - set(non_excludes)
            my_str = self.tokenizer(old_dict, excludes=list(excludes), escape=True)
            new_dict = self.parser(my_str)

            old_keys = set(old_dict.keys())
            new_keys = set(new_dict.keys())
            self.assertEqual(old_keys - excludes, new_keys)
            for k in (new_keys):
                self.assertEqual(old_dict[k], new_dict[k])

        self.assertGreater(c, 0)

    def test_repeated_token_raises_warning(self):
        with warnings.catch_warnings(record=True) as warns:
            # manually reset warning registry before test
            for c, (name, inp) in enumerate(self.repeated):
                plastid.util.services.exceptions.pl_once_registry = {}
                warnings.simplefilter("always")
                msg = "%s parser failed to raise warning for %s" % (self.fmt, name)
                self.parser(inp)
                self.assertEqual(len(warns), c + 1)


@attr(test="unit")
class TestGTF2_TokenParsing(TestGFF3_TokenParsing):
    """Test GTF2 feature reading and writing.
  
    See http://mblab.wustl.edu/GTF22.html
    """

    @classmethod
    def setUpClass(cls):
        cls.fmt = "GTF2"
        cls.tokens = _GTF2_TOKENS
        cls.token_dicts = _GTF2_TOKEN_DICTS
        cls.repeated = _GTF2_REPEATED
        cls.tokenizer = staticmethod(make_GTF2_tokens)
        cls.parser = staticmethod(parse_GTF2_tokens)

    @staticmethod
    def modify_dict(my_dict):
        """Add *gene_id* and *transcript_id* attributes to attribute dictionary,
        because these are required by GTF2 elements
         
        Parameters
        ----------
        my_dict : dict
            Key-value pairs to export
         
        Returns
        -------
        dict
        """
        my_dict = copy.deepcopy(my_dict)
        my_dict["gene_id"] = "some_gene"
        my_dict["transcript_id"] = "some_transcript"
        return my_dict

    def test_token_write_all(self):
        my_dicts = [self.modify_dict(X) for X in self.token_dicts]
        super(TestGTF2_TokenParsing, self).test_token_write_all(dicts=my_dicts)

    def test_token_write_subset(self):
        my_dicts = [self.modify_dict(X) for X in self.token_dicts]
        super(TestGTF2_TokenParsing, self).test_token_write_subset(
            dicts=my_dicts, non_excludes=["gene_id", "transcript_id"]
        )

    @catch_warnings(
        "ignore"
    )  # need to catch warnings because duplicate GTF2 tags raise FileFormatWarnings
    def test_gtf2_list_duplicate_tags(self):
        test_cases = [
            (
                'gene_id "mygene"; transcript_id "mytranscript"; tag "tag value";', {
                    'gene_id': 'mygene',
                    'tag': 'tag value',
                    'transcript_id': 'mytranscript'
                }
            ),
            (
                'gene_id "mygene"; transcript_id "mytranscript"; tag "tag value"; tag "tag value 2";',
                {
                    'gene_id': 'mygene',
                    'tag': 'tag value,tag value 2',
                    'transcript_id': 'mytranscript'
                }
            ),
            (
                'gene_id "mygene"; tag "tag value"; transcript_id "mytranscript"; tag "tag value 2";',
                {
                    'gene_id': 'mygene',
                    'tag': 'tag value,tag value 2',
                    'transcript_id': 'mytranscript'
                }
            ),
        ]
        for inp, outp in test_cases:
            self.assertDictEqual(self.parser(inp), outp)



#!/usr/bin/env python
"""Fetch vectors of :term:`counts` at each nucleotide position in one or more
regions of interest (ROIs).


Output files
------------
Vectors are saved as individual line-delimited files -- one position per line --
in a user-specified output folder. Each file is named for the ROI to which it
corresponds. If a :term:`mask file` -- e.g. from  :py:mod:`~plastid.bin.crossmap`
-- is provided, masked positions will be have value `nan` in output.
"""
import argparse
import inspect
import os
import warnings
import sys
import numpy

from plastid.util.scriptlib.argparsers import (
    AlignmentParser,
    AnnotationParser,
    MaskParser,
    BaseParser,
)
from plastid.util.io.openers import get_short_name
from plastid.util.io.filters import NameDateWriter
from plastid.util.scriptlib.help_formatters import format_module_docstring

warnings.simplefilter("once")
printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))


def main(args=sys.argv[1:]):
    """Command-line program
    
    Parameters
    ----------
    argv : list, optional
        A list of command-line arguments, which will be processed
        as if the script were called from the command line if
        :func:`main` is called directly.

        Default: `sys.argv[1:]`. The command-line arguments, if the script is
        invoked from the command line
    """
    al = AlignmentParser()
    an = AnnotationParser()
    mp = MaskParser()
    bp = BaseParser()

    alignment_file_parser = al.get_parser(conflict_handler="resolve")
    annotation_file_parser = an.get_parser(conflict_handler="resolve")
    mask_file_parser = mp.get_parser()
    base_parser = bp.get_parser()

    parser = argparse.ArgumentParser(
        description      = format_module_docstring(__doc__),
        formatter_class  = argparse.RawDescriptionHelpFormatter,
        conflict_handler = "resolve",
        parents = [base_parser, alignment_file_parser, annotation_file_parser, mask_file_parser]
    ) # yapf: disable

    parser.add_argument("out_folder", type=str, help="Folder in which to save output vectors")
    parser.add_argument(
        "--out_prefix",
        default="",
        type=str,
        help="Prefix to prepend to output files (default: no prefix)"
    )
    parser.add_argument(
        "--format",
        default="%.8f",
        type=str,
        help=r"printf-style format string for output (default: '%%.8f')"
    )
    args = parser.parse_args(args)
    bp.get_base_ops_from_args(args)

    # if output folder doesn't exist, create it
    if not os.path.isdir(args.out_folder):
        os.mkdir(args.out_folder)

    # parse args
    # yapf: disable
    ga          = al.get_genome_array_from_args(args,  printer=printer)
    transcripts = an.get_segmentchains_from_args(args, printer=printer)
    mask_hash   = mp.get_genome_hash_from_args(args,   printer=printer)
    # yapf: enable

    # evaluate
    for n, tx in enumerate(transcripts):
        if n % 1000 == 0:
            printer.write("Processed %s regions of interest" % n)
        filename = "%s%s.txt" % (args.out_prefix, tx.get_name())
        full_filename = os.path.join(args.out_folder, filename)

        # mask out overlapping masked regions
        overlapping = mask_hash.get_overlapping_features(tx)
        for feature in overlapping:
            tx.add_masks(*feature.segments)

        count_vec = tx.get_masked_counts(ga)
        numpy.savetxt(full_filename, count_vec, fmt=args.format)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Utilities for manipulating colors and converting between representations
as RGB hex strings, tuples of floats from 0.0 to 1.0, and tuples of ints
from 0.0 to 255.
"""
import numpy
from matplotlib.colors import colorConverter

process_black = "#222222"


def get_rgb255(inp):
    """Fetch `r, g, b` values where `r, g, b` are integers ranging from 0 to 255

    Parameters
    ----------
    inp : *RGB* or *RGBA* sequence or str
        Can be color hex string,  like "#007ADF",  a matplotlib color letter like "r",
        a matplotlib color name like "black,  et c."
        See :meth:`matplotlib.colors.colorConverter.to_rgb`

    Returns
    -------
    :class:`numpy.ndarray`
        Numpy array of `r, g, b` tuples where `r, g, b` take integer values from 0 to 255
    """
    data = colorConverter.to_rgba_array(inp)
    data = (255 * data[:, :3]).round().astype(int).ravel()

    return data


def get_str_from_rgb(inp):
    """Converts RGB tuples of floats from between 0.0 and 1.0 to RGB hex strings of type #RRGGBB

    Parameters
    ----------
    input : tuple
        Tuple of r, g, b values in range from 0.0 to 1.0

    Returns
    -------
    str
        RGB hex string of form '#NNNNNN'

    Raises
    ------
    ValueError if values are out of range
    """
    return get_str_from_rgb255((255 * numpy.array(inp)).round().astype(int))


def get_str_from_rgb255(inp):
    """Converts RGB tuples of ints from between 0 and 255 to RGB hex strings of type `#RRGGBB`

    Parameters
    ----------
    input : tuple
        Tuple of r, g, b values in range from 0 to 255


    Returns
    -------
    str
        RGB hex string of form '#NNNNNN'

    Raises
    ------
    ValueError if values are out of range
    """
    err_msg = "Cannot convert malformed tuple to string. Should be (0..255, 0..255, 0..255). Is: (%s)" % ", ".join(
        [str(X) for X in inp]
    )
    try:
        assert len(inp) == 3
        for x in inp:
            assert x >= 0 and x <= 255
        stmp = "#{:0>2X}{:0>2X}{:0>2X}".format(*inp)
        assert len(stmp) == 7
        return stmp
    except IndexError:
        raise ValueError(err_msg)
    except AssertionError:
        raise ValueError(err_msg)
    except TypeError:
        raise ValueError(err_msg)


def lighten(data, amt=0.10, is255=False):
    """Lighten a vector of colors by fraction `amt` of remaining possible intensity.

    New colors are calculated as::

        >>> new_colors = data + amt*(1.0 - data)
        >>> new_colors[:, -1] = 1 # keep all alpha at 1.0

    Parameters
    ----------
    data : matplotlib colorspec or sequence of colorspecs
        input color(s)

    amt : float,  optional
        Percentage by which to lighten `r`,  `g`,  and `b`. `a` remains unchanged
        (Default: 0.10)

    is255 : bool,  optional
        If `True`,  rgb values in `data` are assumed to be tween 0 and 255
        rather than 0.0 and 1.0. In this case,  return values will also
        be between 0 and 255.

    Returns
    -------
    numpy.ndarray
        Lightened version of data
    """
    data = colorConverter.to_rgba_array(data)

    new_colors = data + amt * (1.0 - data)
    if is255:
        new_colors = (255 * new_colors).round()

    new_colors[:, -1] = data[:, -1]

    return new_colors


def darken(data, amt=0.10, is255=False):
    """Darken a vector of colors by fraction `amt` of current intensity.

    Parameters
    ----------
    data : matplotlib colorspec or sequence of colorspecs
        input color(s)

    amt : float,  optional
        Percentage by which to darken `r`,  `g`,  and `b`. `a` remains unchanged
        (Default: 0.10)

    Returns
    -------
    numpy.ndarray
        Lightened version of data
    """
    data = colorConverter.to_rgba_array(data)

    new_colors = (1.0 - amt) * data
    if is255:
        new_colors = (255 * new_colors).round()

    new_colors[:, -1] = data[:, -1]

    return new_colors



#!/usr/bin/env python
"""Tests for data structures defined in :py:mod:`plastid.genomics.genome_hash`
"""
import unittest
from random import shuffle
from pkg_resources import resource_filename, cleanup_resources
from nose.plugins.attrib import attr

from plastid.genomics.roitools import Transcript
from plastid.readers.bed import BED_Reader
from plastid.genomics.genome_hash import GenomeHash, BigBedGenomeHash, TabixGenomeHash
from plastid.util.services.decorators import skip_if_abstract
from plastid.util.io.filters import CommentReader
from plastid.readers.bigbed import BigBedReader

from plastid.test.ref_files import REF_FILES
import warnings
warnings.simplefilter("ignore", DeprecationWarning)

#===============================================================================
# INDEX: Test case for GenomeHash
#===============================================================================


def tearDownModule():
    """Remove test dataset files after unit tests are complete"""
    cleanup_resources()


def _name_sort(x):
    return str(x)


class AbstractGenomeHashHelper(unittest.TestCase):
    @skip_if_abstract
    def test_getitem_genomicsegment(self):
        #        """test getitem
        #
        #        1.  Make sure each feature can fetch its own subregion from its own neighborhood
        #
        #        2.  Make sure each feature cannot fetch its own antisense subregion
        #
        #        3.  Make sure that a 5' UTR cannot fetch its corresponding CDS
        #        """
        # make sure we can fetch each transcript's own CDS
        u = 0
        for txid, tx in self.tx_dict.items():
            ol_features = list(self.cds_hash[tx.spanning_segment])
            self.assertIn(
                txid, (X.get_name() for X in ol_features),
                msg="%s failed to fetch its own CDS on same strand. Got %s" % (txid, ol_features)
            )

            # make sure 5' UTR does not fetch CDS
            utr5 = tx.get_utr5()
            if len(utr5) > 0:
                u += 1
                ol_features = list(self.cds_hash[utr5.spanning_segment])
                self.assertNotIn(
                    txid, (X.get_name() for X in ol_features),
                    msg="%s 5' UTR fetched CDS on same strand" % txid
                )

                ol_features = list(self.as_cds_hash[utr5.spanning_segment])
                self.assertNotIn(
                    txid, (X.get_name() for X in ol_features),
                    msg="%s 5' UTR fetched CDS on opposite strand" % txid
                )

            self.assertGreater(u, 0)

        # make sure we don't fetch each transcript's own antisense CDS
        # on opposite strand
        for txid, tx in self.tx_dict.items():
            ol_features = list(self.as_cds_hash[tx.spanning_segment])
            self.assertNotIn(
                txid, (X.get_name() for X in ol_features),
                msg="%s fetched its own name on wrong strand!" % txid
            )

    @skip_if_abstract
    def test_getitem_segmentchain(self):
        #        """test getitem
        #
        #        1.  Make sure each feature can fetch its own subregion from its own neighborhood
        #
        #        2.  Make sure each feature cannot fetch its own antisense subregion
        #
        #        3.  Make sure that a 5' UTR cannot fetch its corresponding CDS
        #        """
        # make sure we can fetch each transcript's own CDS
        u = 0
        for txid, tx in self.tx_dict.items():
            ol_features = list(self.cds_hash[tx])
            self.assertIn(
                txid, (X.get_name() for X in ol_features),
                msg="%s failed to fetch its own CDS on same strand. Got %s" % (txid, ol_features)
            )

            # make sure 5' UTR does not fetch CDS
            utr5 = tx.get_utr5()
            if len(utr5) > 0:
                u += 1
                ol_features = list(self.cds_hash[utr5])
                self.assertNotIn(
                    txid, (X.get_name() for X in ol_features),
                    msg="%s 5' UTR fetched CDS on same strand" % txid
                )

                ol_features = list(self.as_cds_hash[utr5])
                self.assertNotIn(
                    txid, (X.get_name() for X in ol_features),
                    msg="%s 5' UTR fetched CDS on opposite strand" % txid
                )

            self.assertGreater(u, 0)

        # make sure we don't fetch each transcript's own antisense CDS
        # on opposite strand
        for txid, tx in self.tx_dict.items():
            ol_features = list(self.as_cds_hash[tx])
            self.assertNotIn(
                txid, (X.get_name() for X in ol_features),
                msg="%s fetched its own name on wrong strand!" % txid
            )

    @skip_if_abstract
    def test_overlap_stranded(self):
        #        """test get_overlapping_features with strand
        #
        #        1.  Make sure each feature can fetch its own subregion from its own neighborhood
        #
        #        2.  Make sure each feature cannot fetch its own antisense subregion
        #
        #        3.  Make sure that a 5' UTR cannot fetch its corresponding CDS
        #        """
        # make sure we can fetch each transcript's own CDS
        u = 0
        for txid, tx in self.tx_dict.items():
            ol_features = list(self.cds_hash.get_overlapping_features(tx, stranded=True))
            self.assertIn(
                txid, (X.get_name() for X in ol_features),
                msg="%s failed to fetch its own CDS on same strand. Got %s" % (txid, ol_features)
            )

            # make sure 5' UTR does not fetch CDS
            utr5 = tx.get_utr5()
            if len(utr5) > 0:
                u += 1
                ol_features = list(self.cds_hash.get_overlapping_features(utr5, stranded=False))
                self.assertNotIn(
                    txid, (X.get_name() for X in ol_features),
                    msg="%s 5' UTR fetched CDS on same strand" % txid
                )

                ol_features = list(self.as_cds_hash.get_overlapping_features(utr5, stranded=False))
                self.assertNotIn(
                    txid, (X.get_name() for X in ol_features),
                    msg="%s 5' UTR fetched CDS on opposite strand" % txid
                )

            self.assertGreater(u, 0)

        # make sure we don't fetch each transcript's own antisense CDS
        # on opposite strand
        for txid, tx in self.tx_dict.items():
            ol_features = list(self.as_cds_hash.get_overlapping_features(tx, stranded=True))
            self.assertNotIn(
                txid, (X.get_name() for X in ol_features),
                msg="%s fetched its own name on wrong strand!" % txid
            )

    @skip_if_abstract
    def test_overlap_unstranded(self):
        #        """test get_overlapping_features unstranded
        #
        #        1.  Make sure each feature can fetch its own subregion from its own neighborhood
        #
        #        2.  Make sure each feature can fetch its own antisense subregion
        #
        #        3.  Make sure that a 5' UTR cannot fetch its corresponding CDS
        #        """
        u = 0
        for txid, tx in self.tx_dict.items():
            # make sure we can fetch each transcript's own CDS from both sense and antisense
            ol_features = list(self.cds_hash.get_overlapping_features(tx, stranded=False))
            self.assertIn(
                txid, (X.get_name() for X in ol_features),
                msg="%s failed to fetch its own name on same strand." % txid
            )

            ol_features = list(self.as_cds_hash.get_overlapping_features(tx, stranded=False))
            self.assertIn(
                txid, (X.get_name() for X in ol_features),
                msg="%s failed to fetch its own name on opposite strand." % txid
            )

            utr5 = tx.get_utr5()
            if len(utr5) > 0:
                u += 1
                ol_features = list(self.cds_hash.get_overlapping_features(utr5, stranded=False))
                self.assertNotIn(
                    txid, (X.get_name() for X in ol_features),
                    msg="%s 5' UTR fetched CDS on same strand" % txid
                )

                ol_features = list(self.as_cds_hash.get_overlapping_features(utr5, stranded=False))
                self.assertNotIn(
                    txid, (X.get_name() for X in ol_features),
                    msg="%s 5' UTR fetched CDS on opposite strand" % txid
                )

            self.assertGreater(u, 0)


@attr(test="unit")
class TestGenomeHash(AbstractGenomeHashHelper):
    """Test case for :py:class:`GenomeHash`"""

    @classmethod
    def setUpClass(cls):
        """Set up test data for `TestGenomeHash`"""
        cls.binsize = 10000

        with open(REF_FILES["100transcripts_bed"]) as fh:
            cls.transcripts = list(BED_Reader(CommentReader(fh), return_type=Transcript))

        with open(REF_FILES["100cds_bed"]) as fh:
            cls.coding_regions = list(BED_Reader(CommentReader(fh), return_type=Transcript))

        with open(REF_FILES["100cds_antisense_bed"]) as fh:
            cls.coding_antisense = list(BED_Reader(CommentReader(fh), return_type=Transcript))

        cls.tx_dict = {X.get_name(): X for X in cls.transcripts}
        cls.cds_dict = {X.get_name(): X for X in cls.coding_regions}
        cls.as_cds_dict = {X.get_name(): X for X in cls.coding_antisense}

        cls.tx_hash = GenomeHash(cls.tx_dict, do_copy=False, binsize=cls.binsize)
        cls.cds_hash = GenomeHash(cls.cds_dict, do_copy=False, binsize=cls.binsize)
        cls.as_cds_hash = GenomeHash(cls.as_cds_dict, do_copy=False, binsize=cls.binsize)

        cls.shuffled_indices = list(range(len(cls.transcripts)))
        shuffle(cls.shuffled_indices)

    def test_genomehash_create_from_dict(self):
        #        """Test creation `GenomeHash`es from dictionaries without loss of information"""
        gh = GenomeHash(self.tx_dict, do_copy=True)

        found = sorted(list(gh.feature_dict.values()), key=_name_sort)
        expected = sorted(list(self.tx_dict.values()), key=_name_sort)
        self.assertEquals(
            len(expected), len(found),
            "Features lost in creation of GenomeHash from dict. Expected %s, found %s." %
            (len(expected), len(found))
        )
        self.assertEquals(
            sorted(list(gh.feature_dict.values()), key=_name_sort),
            sorted(list(self.tx_dict.values()), key=_name_sort),
            "Features lost in creation of GenomeHash from dict"
        )

    def test_genomehash_create_from_list(self):
        #        """Test creation `GenomeHash`es from lists without loss of information"""
        gh = GenomeHash(self.transcripts, do_copy=True)
        found = sorted(list(gh.feature_dict.values()), key=_name_sort)
        expected = sorted(list(self.tx_dict.values()), key=_name_sort)
        self.assertEquals(
            len(expected), len(found),
            "Features lost in creation of GenomeHash from list. Expected %s, found %s." %
            (len(expected), len(found))
        )
        self.assertEquals(
            expected, found,
            "Features lost in creation of GenomeHash from list:\nFirst:\n%s\nSecond:\n%s" %
            (expected, found)
        )

    def test_get_hash_bins(self):
        for ivc in self.transcripts:
            bin_start = ivc.spanning_segment.start // self.binsize
            bin_end = ivc.spanning_segment.end // self.binsize
            my_range = set(range(bin_start, bin_end + 1))
            self.assertTrue(len(my_range) > 0)
            self.assertEquals(my_range, set(self.tx_hash._get_hash_bins(ivc)))

    def test_genomehash_update_from_dict(self):
        #        """Test addition to a `GenomeHash` from a dictionary without loss of features
        #
        #        1. Add features to an empty dictionary
        #
        #        2. Add features to a non-empty dictionary
        #        """
        tuple_sort = lambda x: _name_sort(x[1])
        gh = GenomeHash({}, do_copy=True)
        tx_items = sorted(list(self.tx_dict.items()), key=tuple_sort)
        tx_values = [X[1] for X in tx_items]
        dict1 = dict(tx_items[:50])
        dict2 = dict(tx_items[50:])
        self.assertEqual(len(dict1), 50)
        self.assertEqual(len(dict2), 50)

        # check values, as opposed to items, because keys of feature_dict
        # are unique numerical IDs as opposed to ames
        gh.update(dict1)
        self.assertEquals(
            sorted(list(gh.feature_dict.values()), key=_name_sort), tx_values[:50],
            "Features lost in first update of empty GenomeHash from dict"
        )

        gh.update(dict2)
        self.assertEquals(
            sorted(list(gh.feature_dict.values()), key=_name_sort), tx_values,
            "Features lost in second update of non-empty GenomeHash from dict"
        )

    def test_genomehash_update_from_list(self):
        #        """Test addition to a `GenomeHash` from a list without loss of features
        #
        #        1. Add features to an empty dictionary
        #
        #        2. Add features to a non-empty dictionary
        #        """
        gh = GenomeHash({}, do_copy=True)
        tx_list = sorted(list(self.tx_dict.values()), key=_name_sort)
        self.assertGreater(len(tx_list), 0)

        gh.update(tx_list[:50])
        self.assertEquals(
            sorted(list(gh.feature_dict.values()), key=_name_sort), tx_list[:50],
            "Features lost in update of empty GenomeHash from list"
        )

        gh.update(tx_list[50:])
        self.assertEquals(
            sorted(list(gh.feature_dict.values()), key=_name_sort), tx_list,
            "Features lost in update of non-empty GenomeHash from list"
        )

    def test_nearby_feature_names_stranded(self):
        #        """Test fetching of nearby feature names
        #
        #        1.  Make sure each feature can fetch its own name from its own neighborhood
        #
        #        2.  Make sure each feature cannot fetch its own name from antisense neighorhood
        #
        #        2.  Make sure transcripts that are not on the same chromosome, not on
        #            the same strand, or that are too far apart on the same chromosome
        #            and strand, do not fetch each other.
        #        """
        # make sure we can fetch each transcript's own CDS
        for txid, tx in self.tx_dict.items():
            nearby_names = self.cds_hash.get_nearby_feature_names(tx, stranded=True)
            self.assertIn(txid, nearby_names, msg="%s failed to fetch its own name" % txid)

        # make sure we don't fetch each transcript's own antisense CDS
        # on opposite strand
        for txid, tx in self.tx_dict.items():
            nearby_names = self.as_cds_hash.get_nearby_feature_names(tx, stranded=True)
            self.assertNotIn(
                txid, nearby_names, msg="%s fetched its own name on wrong strand!" % txid
            )

        # make sure that at least one transcript doesn't fetch
        # another's CDS name if we know they are not:
        #     -on same chromosome
        #     -on same strand
        #     -too far on same chromosme and strand
        c = 0  # count how many times we would expect to fetch self
        for idx, (txid, tx) in zip(self.shuffled_indices, sorted(self.tx_dict.items())):
            fake_tx = self.tx_dict[sorted(self.tx_dict.keys())[idx]]
            if fake_tx.spanning_segment.chrom != tx.spanning_segment.chrom or\
               fake_tx.spanning_segment.strand != tx.spanning_segment.strand or\
                (abs(fake_tx.spanning_segment.start - tx.spanning_segment.start) > self.binsize and\
                 abs(fake_tx.spanning_segment.end - tx.spanning_segment.end) > self.binsize):
                nearby_names = self.cds_hash.get_nearby_feature_names(fake_tx)
                self.assertNotIn(txid, nearby_names)
            else:
                c += 1

        # make sure negative test was not trivial
        # i.e. that we tested at least some set of mismatched transcripts
        self.assertNotEqual(c, len(self.tx_dict))

    def test_nearby_feature_names_unstranded(self):
        #        """Test fetching of nearby feature names, disregarding strand
        #
        #        1.  Make sure each feature can fetch its own name from its own sense neighborhood
        #
        #        2.  Make sure each feature can fetch its own name from antisense neighorhood
        #
        #        2.  Make sure transcripts that are not on the same chromosome, or that are too
        #            far apart on the same chromosome, do not fetch each other.
        #        """
        # make sure we can fetch each transcript's own CDS from both sense asd antisense
        for txid, tx in self.tx_dict.items():
            nearby_names = self.cds_hash.get_nearby_feature_names(tx, stranded=False)
            self.assertIn(
                txid,
                nearby_names,
                msg="%s failed to fetch its own name on same strand: %s" %
                (txid, ",".join(nearby_names))
            )

            nearby_names = self.as_cds_hash.get_nearby_feature_names(tx, stranded=False)
            self.assertIn(
                txid,
                nearby_names,
                msg="%s failed to fetch its own name on opposite strand: %s" %
                (txid, ",".join(nearby_names))
            )

        # make sure that at least one transcript doesn't fetch
        # another's CDS name if we know they are not:
        #     -on same chromosome
        #     -too far on same chromosme and strand
        c = 0  # count how many times we would expect to fetch self
        for idx, (txid, tx) in zip(self.shuffled_indices, sorted(self.tx_dict.items())):
            fake_tx = self.tx_dict[sorted(self.tx_dict.keys())[idx]]
            if fake_tx.spanning_segment.chrom != tx.spanning_segment.chrom or\
                (abs(fake_tx.spanning_segment.start - tx.spanning_segment.start) > self.binsize and\
                 abs(fake_tx.spanning_segment.end - tx.spanning_segment.end) > self.binsize):
                nearby_names = self.cds_hash.get_nearby_feature_names(fake_tx)
                self.assertNotIn(txid, nearby_names)
            else:
                c += 1

        # make sure negative test was not trivial
        # i.e. that we tested at least some set of mismatched transcripts
        self.assertNotEqual(c, len(self.tx_dict))

    def test_nearby_features_stranded(self):
        #        """Test fetching of nearby features
        #
        #        1.  Make sure each feature can fetch its own subregion from its own neighborhood
        #
        #        2.  Make sure each feature cannot fetch its own antisense subregion
        #
        #        3.  Make sure fatures that are not on the same chromosome, not on
        #            the same strand, or that are too far apart on the same chromosome
        #            and strand, do not fetch each other.
        #        """
        # make sure we can fetch each transcript's own CDS
        for txid, tx in self.tx_dict.items():
            nearby_features = self.cds_hash.get_nearby_features(tx)
            self.assertIn(
                txid, (X.get_name() for X in nearby_features),
                msg="%s failed to fetch its own CDS on correct strand" % txid
            )

        # make sure we don't fetch each transcript's own antisense CDS
        # on opposite strand
        for txid, tx in self.tx_dict.items():
            nearby_features = self.as_cds_hash.get_nearby_features(tx, stranded=True)
            self.assertNotIn(
                txid, (X.get_name() for X in nearby_features),
                msg="%s fetched its own name on wrong strand!" % txid
            )

        # make sure that at least one transcript doesn't fetch
        # another's CDS if we know they are not:
        #     -on same chromosome
        #     -on same strand
        #     -too far on same chromosme and strand
        c = 0  # count how many times we would expect to fetch self
        for idx, (txid, tx) in zip(self.shuffled_indices, sorted(self.tx_dict.items())):
            fake_tx = self.tx_dict[sorted(self.tx_dict.keys())[idx]]
            if fake_tx.spanning_segment.chrom != tx.spanning_segment.chrom or\
               fake_tx.spanning_segment.strand != tx.spanning_segment.strand or\
                (abs(fake_tx.spanning_segment.start - tx.spanning_segment.start) > self.binsize and\
                 abs(fake_tx.spanning_segment.end - tx.spanning_segment.end) > self.binsize):
                nearby_features = self.cds_hash.get_nearby_features(fake_tx)
                self.assertNotIn(txid, (X.get_name() for X in nearby_features))
            else:
                c += 1

        # make sure negative test was not trivial
        # i.e. that we tested at least some set of mismatched transcripts
        self.assertNotEqual(c, len(self.tx_dict))

    def test_nearby_feature_unstranded(self):
        #        """Test fetching of nearby features, disregarding strand
        #
        #        1.  Make sure each feature can fetch its own subregion from its own neighborhood
        #
        #        2.  Make sure each feature can fetch its own antisense subregion
        #
        #        3.  Make sure fatures that are not on the same chromosome, or that are too far
        #            apart on the same chromosome, do not fetch each other.
        #        """
        # make sure we can fetch each transcript's own CDS from both sense asd antisense
        for txid, tx in self.tx_dict.items():
            nearby_features = self.cds_hash.get_nearby_features(tx, stranded=False)
            self.assertIn(
                txid, (X.get_name() for X in nearby_features),
                msg="%s failed to fetch its own name on same strand." % txid
            )

            nearby_features = self.as_cds_hash.get_nearby_features(tx, stranded=False)
            self.assertIn(
                txid, (X.get_name() for X in nearby_features),
                msg="%s failed to fetch its own name on opposite strand." % txid
            )

        # make sure that at least one transcript doesn't fetch
        # another's CDS if we know they are not:
        #     -on same chromosome
        #     -too far on same chromosme and strand
        c = 0  # count how many times we would expect to fetch self
        for idx, (txid, tx) in zip(self.shuffled_indices, sorted(self.tx_dict.items())):
            fake_tx = self.tx_dict[sorted(self.tx_dict.keys())[idx]]
            if fake_tx.spanning_segment.chrom != tx.spanning_segment.chrom or\
                (abs(fake_tx.spanning_segment.start - tx.spanning_segment.start) > self.binsize and\
                 abs(fake_tx.spanning_segment.end - tx.spanning_segment.end) > self.binsize):
                nearby_features = self.cds_hash.get_nearby_features(fake_tx, stranded=False)
                self.assertNotIn(txid, (X.get_name() for X in nearby_features))
            else:
                c += 1


@attr(test="unit")
class TestBigBedGenomeHash(AbstractGenomeHashHelper):
    @classmethod
    def setUpClass(cls):
        """Set up test data for `TestGenomeHash`"""
        cls.binsize = 10000

        cls.tx_bbfile = REF_FILES["100transcripts_bigbed"]
        cls.cds_bbfile = REF_FILES["100cds_bigbed"]
        cls.as_cds_bbfile = REF_FILES["100cds_antisense_bigbed"]

        cls.tx_hash = BigBedGenomeHash(cls.tx_bbfile)
        cls.cds_hash = BigBedGenomeHash(cls.cds_bbfile)
        cls.as_cds_hash = BigBedGenomeHash(cls.as_cds_bbfile)

        cls.transcripts = list(BigBedReader(cls.tx_bbfile, return_type=Transcript))
        cls.coding_regions = list(BigBedReader(cls.cds_bbfile))
        cls.shuffled_indices = list(range(len(cls.transcripts)))

        cls.tx_dict = {X.get_name(): X for X in cls.transcripts}
        cls.cds_dict = {X.get_name(): X for X in cls.coding_regions}
        shuffle(cls.shuffled_indices)

    def test_works_with_str_filename(self):
        one_hash = BigBedGenomeHash(self.tx_bbfile)
        for tx in self.transcripts:
            expected = self.tx_hash[tx]
            found = one_hash[tx]
            self.assertEqual(expected, found)

    def test_works_with_multiple_files(self):
        # check to make sure all files are queried by comparing results
        # from a single hash to results from a double hash
        double_hash = BigBedGenomeHash(self.tx_bbfile, self.cds_bbfile)
        for tx in self.transcripts:
            expected = self.tx_hash[tx] + self.cds_hash[tx]
            found = double_hash[tx]
            self.assertEqual(expected, found)


@attr(test="unit")
class TestTabixGenomeHash(AbstractGenomeHashHelper):
    @classmethod
    def setUpClass(cls):
        """Set up test data for `TestTabixGenomeHash`"""

        cls.tx_file = REF_FILES["100transcripts_bed_tabix"]
        cls.cds_file = REF_FILES["100cds_bed_tabix"]
        cls.as_cds_file = REF_FILES["100as_cds_bed_tabix"]

        cls.tx_hash = TabixGenomeHash(cls.tx_file, data_format="BED")
        cls.cds_hash = TabixGenomeHash(cls.cds_file, data_format="BED")
        cls.as_cds_hash = TabixGenomeHash(cls.as_cds_file, data_format="BED")

        # use BigBeds as reference
        # TODO: change to Transcript objects
        cls.tx_bbfile = REF_FILES["100transcripts_bigbed"]
        cls.cds_bbfile = REF_FILES["100cds_bigbed"]
        cls.as_cds_bbfile = REF_FILES["100cds_antisense_bigbed"]

        cls.transcripts = list(BigBedReader(cls.tx_bbfile, return_type=Transcript))
        cls.coding_regions = list(BigBedReader(cls.cds_bbfile))
        cls.shuffled_indices = list(range(len(cls.transcripts)))

        cls.tx_dict = {X.get_name(): X for X in cls.transcripts}
        cls.cds_dict = {X.get_name(): X for X in cls.coding_regions}
        shuffle(cls.shuffled_indices)

    def test_unknown_reader_raises_error(self):
        self.assertRaises(KeyError, TabixGenomeHash, self.tx_file, data_format="GARBAGE")

    def test_works_with_str_filename(self):
        one_hash = TabixGenomeHash(self.tx_file, data_format="BED")
        for tx in self.tx_dict.values():
            found = one_hash[tx]
            expected = self.tx_hash[tx]
            self.assertEqual(expected, found)

    def test_works_with_multiple_files(self):
        # check to make sure all files are queried by comparing results
        # from a single hash to results from a double hash
        double_hash = TabixGenomeHash(self.tx_file, self.cds_file, data_format="BED")
        for tx in self.tx_dict.values():
            df = double_hash[tx]
            expected = self.tx_hash[tx] + self.cds_hash[tx]
            self.assertEqual(len(df), len(expected))

    def test_works_with_multiple_files_list(self):
        # check to make sure all files are queried by comparing results
        # from a single hash to results from a double hash
        double_hash = TabixGenomeHash([self.tx_file, self.cds_file], data_format="BED")
        for tx in self.tx_dict.values():
            df = double_hash[tx]
            expected = self.tx_hash[tx] + self.cds_hash[tx]
            self.assertEqual(len(df), len(expected))



#!/usr/bin/env python
"""Unit tests for functions in :py:mod:`plastid.bin.slidejuncs`
"""

from nose.plugins.attrib import attr
from nose.tools import assert_equal, assert_set_equal, assert_greater
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from plastid.bin.slidejuncs import (
    find_match_range,
    find_known_in_range,
    find_canonicals_in_range,
    covered_by_repetitive,
)
from plastid.genomics.roitools import SegmentChain, GenomicSegment
from plastid.genomics.genome_hash import GenomeHash

#===============================================================================
# INDEX: verify test data above in case it got bungled
#===============================================================================


def check_data_by_sequence(expected_ivc, test_ivc):
    """Verify that two junctions produce identical sequence
    
    Parameters
    ----------
    expected_ivc : SegmentChain
        SegmentChain corresponding to known splice junction
    
    test_ivc : SegmentChain
        SegmentChain corresponding to test splice junction that is sequence-consistant
        with ``expected_ivc``
    """
    expected_seq = expected_ivc.get_sequence(seqs)
    test_seq = test_ivc.get_sequence(seqs)
    assert_equal(
        expected_seq, test_seq, "Sequences don't match in test dataset for %s and %s." %
        (expected_seq.get_name(), test_seq.get_name())
    )


def test_data_by_sequence():
    for txid in seqs:
        expected_ivc = known_juncs.get(txid[0], None)
        if expected_ivc is not None:
            for query_ivc in query_juncs[txid]:
                yield check_data_by_sequence, expected_ivc, query_ivc


#===============================================================================
# INDEX: testing match range
#===============================================================================


def check_find_match_range(txid, max_slide):
    """Check output of :py:func:`find_match_range` against known data
    
    Parameters
    ----------
    txid : str
        ID of transcript to check
    """
    my_juncs = known_juncs.get(txid, [])
    for n, my_junc in enumerate(my_juncs):
        ref_minus, ref_plus = match_ranges[txid][n]
        ref_minus = max(-max_slide, ref_minus)
        ref_plus = min(max_slide + 1, ref_plus)
        found_minus, found_plus = find_match_range(my_junc, seqs, max_slide)
        assert_equal(
            ref_minus, found_minus,
            "Minus match ranges don't match. Expected %s, got %s" % (ref_minus, found_minus)
        )
        assert_equal(
            ref_plus, found_plus,
            "Plus match ranges don't match. Expected %s, got %s" % (ref_plus, found_plus)
        )


@attr(test="unit")
def test_find_match_range():
    # test case, plus strand has match range plus
    # test case, plus strand has no match range plus
    # test case, minus strand has match range plus
    # test case, minus strand has no match range plus
    for txid in seqs:
        for max_slide in (2, 4, 20):
            yield check_find_match_range, txid, max_slide


#===============================================================================
# INDEX: find known junctions within match range only when present
#===============================================================================


def check_find_known_in_range(query_junc):
    txid = query_junc.chrom
    minus_range, plus_range = find_match_range(query_junc, seqs, 20)
    found_known = find_known_in_range(query_junc, minus_range, plus_range, all_known_juncs)

    expected = set([str(X) for X in known_juncs.get(txid, [])])
    found_known = set([str(X) for X in found_known])
    assert_set_equal(expected,found_known,("Known junction not found for %s. "+\
                                           "Query: %s\n    Expected %s\n    " +\
                                           "Got %s\n    match range: (%s,%s)") % (txid,
                                                                                str(query_junc),
                                                                                expected,
                                                                                found_known,
                                                                                minus_range,
                                                                                plus_range))


@attr(test="unit")
def test_find_known_in_range():
    # test case, has known in plus range plus strand
    for txid in seqs:
        juncs = query_juncs.get(txid, [])
        for my_junc in juncs:
            yield check_find_known_in_range, my_junc


def check_find_none_when_none_known_in_range(query_junc):
    minus_range, plus_range = find_match_range(query_junc, seqs, 20)
    found_known = find_known_in_range(query_junc, minus_range, plus_range, all_known_juncs)
    found_known = set([str(X) for X in found_known])
    assert_set_equal(found_known, set([]))


@attr(test="unit")
def test_find_none_when_none_known_in_range():
    for _, v in unmatched_query_juncs.items():
        for junc in v:
            yield check_find_none_when_none_known_in_range, junc


#===============================================================================
# INDEX: find canonical junctions within match range only when present
#===============================================================================


def check_find_canonicals_in_range(query_junc):
    txid = query_junc.chrom
    minus_range, plus_range = find_match_range(query_junc, seqs, 20)
    found_can = find_canonicals_in_range(
        query_junc, minus_range, plus_range, seqs, canonicals[query_junc.strand]
    )
    found_can = set([str(X) for X in found_can])
    expected = set([str(X) for X in known_juncs.get(txid, [])])
    assert_set_equal(expected,found_can,("Known junction not found for %s. "+\
                                           "Query: %s\n    Expected %s\n    " +\
                                           "Got %s\n    match range: (%s,%s)") % (txid,
                                                                                str(query_junc),
                                                                                expected,
                                                                                found_can,
                                                                                minus_range,
                                                                                plus_range))


@attr(test="unit")
def test_find_canonicals_in_range():
    for txid in seqs:
        juncs = noncan_juncs.get(txid, [])
        for my_junc in juncs:
            yield check_find_canonicals_in_range, my_junc


def check_find_none_when_none_canonical_in_range(query_junc):
    minus_range, plus_range = find_match_range(query_junc, seqs, 20)
    found_known = find_canonicals_in_range(
        query_junc, minus_range, plus_range, seqs, canonicals[query_junc.strand]
    )
    found_known = set([str(X) for X in found_known])
    assert_set_equal(found_known, set([]))


@attr(test="unit")
def test_find_none_when_none_canonical_in_range():
    for junc in unmatched_noncan_query_juncs:
        yield check_find_none_when_none_canonical_in_range, junc


#===============================================================================
# INDEX: find repetitive junctions within match range only when present
#===============================================================================


@attr(test="unit")
def test_find_repetitive_in_range():
    pos = 0
    neg = 0
    # some should be positive, others negative
    for txid in known_juncs:
        expected = txid in cross_hash_seqs
        if expected == True:
            pos += 1
        else:
            neg += 1
        my_query_juncs = query_juncs.get(txid, [])
        for query_junc in my_query_juncs:
            minus_range, plus_range = find_match_range(query_junc, seqs, 20)
            yield check_find_repetitive_in_range, query_junc, minus_range, plus_range, expected

    # all negative
    for txid in unmatched_query_juncs:
        my_query_juncs = unmatched_query_juncs.get(txid, [])
        for query_junc in my_query_juncs:
            minus_range, plus_range = find_match_range(query_junc, seqs, 20)
            yield check_find_repetitive_in_range, query_junc, minus_range, plus_range, False

    # make sure we found a bunch of each type
    assert_greater(pos, 0)
    assert_greater(neg, 0)


def check_find_repetitive_in_range(query_junc, minus_range, plus_range, expected):
    """Find repetitive blocks that overlap splice junctions
     
    Parameters
    ----------
    query_junc : SegmentChain
        splice junction
     
    expected : bool
        Whether or not ``query_junc`` overlaps a repetitive feature
    """
    assert_equal(covered_by_repetitive(query_junc, minus_range, plus_range, cross_hash), expected)


#===============================================================================
# INDEX: test data for tests above
#===============================================================================

canonicals = {
    "+": [("GT", "AG"), ("GC", "AG")],
    "-": [("CT", "AC"), ("CT", "GC")],
}
"""Canonical splice junction sequences for each strand"""

seqs = {
    "YNL130C_fake":
    "AAAAGGGCCTCAATACATGGTTTGAAAGGAAAGAGCGGTCATCACTTTGATACCTAAAGTCAGATTAGTAATTTAATGTTACAATAGTTAGTATTTATTTTTCTTGCCACAGAAGATTTAGGGTGCAATAAGATAAGCAACATACTTGTAAAGCTTCAAGTTTCCCAAACTACTCTGAGGAATAAAGAAT",
    "YNL130C":
    "AAAAGGGCCTCAATACATGGTTTGAAAGGAAAGAGCGGTCATCACTTTGATACCTAAAGTCAGATTAGTAATTTAATGTTACAATAGTTAGTATTTATTTTTCTTGCCACAGAAGATTTAGGGTGCAATAAGATAAGCAACATACTTGTAAAGCTTCAAGTTTCCCAAACTACTCTGAGGAATAAAGAAT",
    "YPL249C-A":
    "GGCTGGAGTCATTTGGGTGACTTTCTTACCCTTGTTCAAACCAATAGCGATACCTAGAGGAAAAAAATAGAATATCTTATGTTAGTAACAAGCACACTAGTACTCACGTAGTATACGAGCAAAATAAATAAACAAAGACTCCACGGAGTACTGCAACGCTGGGTGCAGGATATAAACGATGTTTGCATATCTCTGGAGCTTAATCTCATCCTGTTTGTATTTCTGTTGTTGGCTTAGTGGTTCAGTTCATTTTGCTTTTGCTCTACAGTTCATTAGTTCAGGTAAACATACCAGTCTTGACAGCCATTTTGTATTATCCTTGCTTCTCTATTCTG",
    'YBR215W_mRNA_0':
    'ACACACAATGCAGAAACTGATGCGGTTAATTTTGCATTCTGAAACGATTTAAACGAACAGCATTAACCTCCACGACCATATTCAAACGATTGGAAATGGACCAAAAAGGTATGTTCTGGCTTTATTTTCAATTATCCGCCCTTCGTTAGGGGTGTTTTAGATCATTTTGATTAACAATTCTATTTATGATAGCAATTGTCCTTGATAACTCTAAAAGTGGTAGTAAACAGACAAAATCGAGTGGCAAAATGCAAACGCAAACAGATACTAATGCTGAAGTTCTAAACACCGATAACAGTATCAAAAAAGAAACAGGAAGTGATTCTGAGGATCTTTTCAATAAATTTTCAAACAAAAAAACGAACAGGAAAATACCCAACATCGCTGAAGAATTAGCGAAAAACAGAAACTACGTAAAAGGTGCGTCTCCGTCTCCCATTATAATTTCTGGTTCCTCTTCAACTTCTCCATCAGGACCCTCTTCCTCCAGTACAAACCCGATGGGGATACCAACGAATCGATTTAATAAAAATACTGTAGAACTATATCAGCATTCACCATCCCCCGTGATGACTACTAACAAGACTGATACAGAGGAAAAGAGACAAAACAATAGAAATATGGATAATAAAAACACCCCAGAAAGAGGATCTTCTTCCTTTGCTGCTAAGCAACTAAAAATATCATCCTTGTTAACTATATCCTCCAATGAAGATAGTAAAACTTTGCATATAAACGATACTAATGGCAACAAAAATAGTAATGCCGCCAGTAATAATATTCCATCGGCTTATGCAGAACTCCATACAGAGGGAAATAGCATTGAATCACTAATAAAACCCCCAAGCTCTCCCAGAAATAAATCTCTAACGCCCAAAGTTATCTTGCCAACACAGAATATGGATGGTACAATTGCAAAGGATCCTCATTTAGGTGATAATACACCAGGAATACTCATAGCAAAAACTAGTTCTCCCGTAAATTTGGATGTTGAAAGCACTGCACAATCTCTGGGAAAATTTAACAAGTCCACTAATTCTTTGAAAGCTGCCCTAACAAAAGCTCCTGCAGAAAAGGTCTCTTTAAAACGTAGCATTAGTTCAGTTACAAACAGTGATTCTAATATTAGCTCCAGCAAAAAGCCTACGTCTGAAAAAGCTAAAAAATCAAGTTCAGCATCAGCAATACTACCAAAGCCAACAACGACCAAGACATCAAAAAAAGCTGCTAGCAATAGTAGCGATTCTACCAGAAAAAAAAACGCCTCGAACAAGACTACATCAGCCATAAAGAAAGAGTCAAATGCCGGCTCCAAATTGAACACTGTTAAGAAGGAAAATAGCTCTTTATCTTCTATCAAAGCCACTGAGAAAGAAAAAGATAAAGGTGGAAATAGCACGGAAGCAAAAAACTCTACCAGCAACGTTAGAAAGGAACCAACTGCAAAATCCCCAAAAAGGTTAGTAGCCGCACCGACAGTCAGTCCACCAAAAATTCTACAAACGGCAGAAACCAAAGCAAAGGAACCCTCGATATTGATTGACGTTCCATTGTATCAAGCTGATACAAACGACTACCTAGACGAAAACGGTCAGGTTATTTTTAATTTATCAACTTTAATAAAAGAGAAGTATCACCCGAAAAGTAAAGAACTAGCACAACTTAAAGACTCTAAAAGAAATTTATTAATGCAATTATCCGATCATTCTAACGGCTCACTAGAAAAGGAAAAAGATGAAGAAGGAGATGTAATAGAACTTGATGATGATGAGGATATGGAAGAAGATGAAGGGGAAATAGATACAGAAACTAATACTGTAACTACCACCATCTCTCCGAAGAAGAAGTCGCATCCTATGAAAGGTAAAAATTTGATTGGTAAATATGACGTTGAAGATCCGTTCATTGATGATTCGGAGTTATTGTGGGAGGAACAGCGTGCTGCCACCAAGGACGGTTTTTTTGTTTACTTTGGTCCATTAATCGAAAAAGGTCACTATGCAAGTTTGGAGCGTGCAAATGGTACCATGAAAAGAGGTGGTGTTAAAAATAAATAAGCATTGTGTCAGAGATTTTCCAAAATCCCTGAACGACATAAAAAATGAAATAAAGTGTGCCTATATCTGAAAGTATATAGTCAGTTTATATGAAGATTTTTCCGAGCGAATGCTTTTC',
    'YHL001W_mRNA_0':
    'GCGCAAATAAACCAAAAATGTCCACTGATTCTATTGTTAAAGCCTCCAACTGGAGATTAGTCGAAGTCGGCCGTGTCGTTTTGATCAAAAAGGGTCAATCCGCAGGTAAATTGGCTGCTATCGTTGAAATTATCGACCAAAAGAAGGTATGTTGAACCTAAAACCCACCGTGGACAAACTGAGGAGGAAATTGTAAGGAAGAGAAAGTCCCCGTATGTTCAGGGACCGCCAACGACTTCCTGATTCATGCCTGAATAGTGAAACATGTTAAACAATGAGTAGATGCTGAAATATGCAATGGGACAGAGGGCCGGAGATGTTTCTATTCTTTTTTTGCATAAACAACAGTGAAATTTCTACATATTTTGTTACTGGCAAACTCCTTCCCTGGCAGTTAATAACTCCCATACTTAAGAAGATAACTTCAGCACATGTAGCCATACCTCCTCAATTGCCAAATAATCTTTTCCTGTTTTCAAAGTATTTCAATTTACTAACAACTTTTATGAATAATTTTTTAAATTATTCGTTATAATTACTATAGGTTTTGATTGATGGTCCAAAAGCTGGTGTCCCACGTCAAGCCATCAACTTGGGTCAAGTTGTCTTAACTCCATTGACCTTTGCTCTACCAAGAGGTGCTAGAACTGCTACCGTTTCTAAGAAGTGGGCTGCTGCTGGTGTTTGCGAAAAGTGGGCTGCTTCATCTTGGGCTAAGAAGATTGCTCAACGTGAAAGACGTGCTGCTTTGACTGACTTTGAAAGATTCCAAGTTATGGTTTTAAGAAAGCAAAAGAGATACACTGTCAAGAAGGCTTTGGCTAAGGCTTAAAAAAAGAAGAATAATTCTAAAATCCATAGGTAAGTACTGAAAGCAATTTTGCGTTCCGTCAATGCATATTATATATATTAATCTTAACCATTTATGTAAACAACATATCATTTCATTTTGTTCTGGCCA',
    'YIL018W_mRNA_0':
    'CCAAGAAACCACAAAGTTATTGAACAATGGGTATGTTTTTATTATATCGCATAATTATGGCAAATGTTATGAAGGATTCCTCTATGACTTAGATGTTTTGAATCGGTACATTTTATTTTCAGTATCCTTCTGCATATTACAAGGCAACATAGCAGCGGACAAGAAAGCTTTTTTGATGTTCGTCTTCGAAACGATTACTATCAGGGTCTTTTAAATGCTATATCGGGACTTATTGAAATTGACATATTACACTTATGAATGACGTTGGCTATCAAAGTATGAGAATAAGCCTTTCCTACTATTTTGTATCATGACAACAGGGTTCTGTCGCTTTGAATGCGGCCTTTTACTTTGCCATATTTTGATAATGAAAAAACTTTCGAGAAATATTTACTAACAGGATCGAGAATTTTTTTCCTCATTTAAACAGGTAGAGTTATTCGTAACCAAAGAAAGGGTGCTGGTTCTATCTTCACCTCTCACACCAGATTAAGACAAGGTGCTGCCAAGTTGAGAACTTTGGACTATGCTGAACGTCATGGTTACATCCGTGGTATCGTTAAGCAAATTGTCCACGACTCCGGTAGAGGTGCTCCATTGGCCAAGGTTGTTTTCCGTGACCCATACAAGTACAGATTACGTGAAGAAATCTTCATTGCTAACGAAGGTGTTCACACTGGTCAATTCATTTACGCTGGTAAAAAGGCTTCTTTGAACGTCGGTAACGTCTTGCCATTGGGTTCTGTCCCAGAAGGTACCATTGTCTCCAACGTTGAAGAAAAGCCAGGTGACAGAGGTGCCCTAGCCAGAGCTTCTGGTAACTACGTTATCATCATTGGTCACAACCCAGATGAAAACAAAACCAGAGTCAGATTACCATCCGGTGCCAAGAAGGTTATCTCTTCTGACGCCAGAGGTGTCATCGGTGTCATTGCCGGTGGTGGTAGAGTTGACAAACCATTGTTGAAGGCTGGTCGTGCTTTCCACAAGTACAGATTGAAGAGAAACTCTTGGCCAAAGACCCGTGGTGTTGCCATGAATCCAGTTGATCACCCTCACGGTGGTGGTAACCATCAACATATTGGTAAGGCTTCTACTATCTCTAGAGGTGCCGTTTCTGGTCAAAAGGCTGGTTTGATTGCTGCCAGAAGAACCGGTTTACTACGTGGTTCTCAAAAGACCCAAGATTAATCTTTTTAATTTTGGTTTCTTCCTTCTGTCATATTATTTTATCAATTTTCTTAAATATTATATAATTTAATCCGAAACGTTCCTTATAA',
    'YIL133C_mRNA_0':
    'TGGTTTTTATTATTTCTAAAAAATATTTGTGATTATATTACAAAATGTAAAGTATTAAAGAAAAACTATTTTTAATTTCTTAGTAACCCAAAGCGGCTAATTGTTTAGCAACATCAGATTCAGCAGCAGTAGCATTAGCAGAGGCAACTTTCTTGGTGAAAGCTCTCTTCTTGGCATAGTATTCAGCGGATGAAACCTTTCTCTTTGCTTCCAATTTGGCAACAACATCTTCGTATTTCCAACCAACGGAAGTAGACAACTTACCCAAGGTGGTGTACTTTCTACCTGGCTTCAATCTCAAAACTCTCAATGCTTGTGGGACAACAACTCTCTTCTTCTTATCGTATGGAGGTGGGATACCTTCAAAGACCTTTAAACGTTCCAAAGCGGCCTTACCACGAGCAGTCTTGTGAGAAACCATACCACGAAGAGCTTTGTAGAAGATTCTAGATGGGGCTCTGAAATGGAATGGACCACGGGTCTTATTGAAAGCGGTAGCCTTTCTCAAGAAATCATGGTACTTTAATTTGTTTCTGAAAAATTCACCAGAAATGTTTAATTCTTCAGCTCTAACAACAACGATTTTTTGACCGTTTAATAATTGCTTAGCAACAACGGAAGCTAAACGACCTACTAAATGACCCTTACCTGGTTGAAGATTATTATTGATGAATGAATGGATGCTCATTCAAAAAAATGGATTTGTTAGTATTATCCTTTAAATAAAACCTGTTTTTCCCGTTCATAATTCCCATCTCTCTCTATCTCCATGTTTTAGCGTATACACAATCTACGAAGACCCTTCTGTATTCCACCCTTCTGCGGCTGTAATATTTTTATCATATTCCTTGGGGACTCGTAAACGCTGTGCACTATGAGCAAATCCTGCGGCTGGGAAAGGAAAACACATTCAATTATTAGTGCAAAATCTTACGTACCATCAATGACAACAACTGGTTCAACAGACATTTTCTCGATTTGTTCTTCACCTTCCGTTTTCAAAGAGA',
    'YIL156W_B_mRNA_0':
    'GATCAAAACAAGAACAAGCAAAACGCAAAGAGATGACATTAGTAAGTACCCAAATGAGCTACTAACAACGCATCCGGTAATACTAACAAGAGAAATTGGTTAGGTAGGGAAACTAGTACACATATCCATAGATTTGGTCTTGGTATCAACGTGCCTCGCGGGGATTAAACGCAATACAGGTTTGACGCCTAAACTGGAGACTCTGGACAATCAGACTATGAGAAACTACATGAAGCGGTACTTGAACTTGGGAGAGTCTGTATACGACTATTCTGTTGCCACTTGTGGCTCCTCGACGTATTTCGCTAGGAAATAGGTACGTTTATATTCTATATATGTTGATAATATTTTGTATAACGGCACCCAGCCGGGGGCCAAAATTAATAACCGCAACTTAAAGAGCTATCG',
    'YKL006W_mRNA_0':
    'AAGAGAGTCGTGAAAAATAAAATAAACAATGTCCACCGATTCTATTGTCAAGGCTTCTAACTGGAGATTAGTCGAAGTTGGCCGTGTGGTTTTGATCAAGAAGGGTCAATCCGCAGGTAAATTGGCCGCTATCGTCGAAATTATTGACCAAAAGAAGGTATGTTAAATCCGGAAAACCTATCATCGATTTGAGGAGGGAAATGAACGAGGAATGAGATTTAGGCTACGAAAAAGTGCTGGCAAATATCCAATCCGTTACCATGGATGAAGCTACTTGTAAACGATAAGTAAATGTGAAAATACACATTGCGATAGAGATACAAAAAAGTCTTTATTTCCCTGTTGGAGTGGCCTATAGTGGATTATTTATTTATTTTGCTAGCAGCTTACCTAGTTTCAGAAAGGTGATGACCCATATTTGCGCAGTTAATTCTTAAAGAGCCTGACCATGTTTCCTCAAATATTAAACCATGCCTTCCATCGTTTTGAAAAAATTTCGTTTACTAACAACTTTTATGAATAATTTTTTTAAATTATTTGTTATAATTATTATAGGTTTTGATTGACGGTCCAAAAGCTGGTGTTCCACGTCAAGCCATCAACTTGGGTCAAGTTGTCTTAACTCCATTGACCTTTGCTCTACCAAGAGGTGCTAGAACTGCTACCGTTTCTAAGAAGTGGGCTGCTGCTGCTGTCTGCGAAAAGTGGGCTGCTTCATCTTGGGCTAAGAAGATTGCTCAACGTGAAAGACGTGCTGCTTTGACTGACTTTGAAAGATTCCAAGTTATGGTTTTAAGAAAGCAAAAGAGATACACTGTCAAGAAGGCTTTGGCTAAGGCTTAAAATATGTACATGATCTTTAATTCTGATATATTTCGTATGTAATTTTATCTTTAACTGGTGATTCTTTTAATAAATAAACTACAATATTATAACTATTGAAAGCCTCTGTTA',
    'YMR194C_B_mRNA_0':
    'TCTTTTTCATAAAGAAAGTGAGTGGATCTTCCCTGGAAGTACGAATAATGTACTTATGATTAGATTCTCTTCACGCACATACCTTTTTTTTTTTTTCAATTGCTAGTTCATACATACGTAGAAAGAAGGGCCAATTGGAAGCCTCAACTGTCCCCTGGTGTAAGCTTTCGCTGTTTCATCTTCAACTCAAGAAGCGAAGGTAGCGGACAGCAGGGTGACCGTGAATCTTTTCCATTGTCATTGTAAAATTTTGAGCAGCAGATGTACAGCTGATCTATAACTTTAGCGCATTTAGCATCATCGTACTGATGTGATAGAAGGCAATCTAAAATTAATGTGAAAAAACGTTAGTGAAAAGAAAAAAACGATCGAATCTTTTGAAGTTAATATTACAAACCTTGGATGGCGCAGGCCTCCTTCTGGCATGGATTAGACATATTTACCTCTTATTTGAACAAATCTCAACAACTAAAAAAACGCTCAATATCTTGGAAAAAAAGAACAGTAATATTCTGTCTCTCCACACCCTCTTCTGAACTCTCATATACACCCCATGGGTGAAAAGTAATATCCGTAAAAGTCTCGCTAACGGATGACGTTAACAACGCACGCCCGCTCCTTGCGAAATTTATTGTGGGGACGTCGTTCACGTGCAAAAGTCGCCAAATATTAGCGACGGGCAACATCGGAAAAACTCTGCGCCCGCGATTCCGATTGAGATGTGTAATT',
    'YPL249C_A_mRNA_0':
    'CGATTAAGAATTACTATAATGATTTGTTACTTATTTTTTATTAACTAGCTTTGGGGGAGAGCCATGGAAAATAGCACTCGGTCTTGTGGCGGAGTTGGATTTGCTTGACTGTGGCCCCCTCACGCTGTTTAATGACGACGAGAGGCAGCAATGATGTTGTTCATTTCTTCGACCTTAGCCTTGGCTCTGGTGAAAGAACCCAATCTCTTCTTGGCGACCTTTCTGGCTCTCTTTTCACCGGAGTTTCTGATCAAATCGATCAATCTTCTTTCATATGGGGACAAACCGGCGATTTCTCTAACCAAAGATCTGACGAACTTGGTTCTGTTGGAGGCAGCACCCTTCTTGTAGGAGATCTTTGGGGCTGGAGTCATTTGGGTGACTTTCTTACCCTTGTTCAAACCAATAGCGATACCTAGAGGAAAAAAATAGAATATCTTATGTTAGTAACAAGCACACTAGTACTCACGTAGTATACGAGCAAAATAAATAAACAAAGACTCCACGGAGTACTGCAACGCTGGGTGCAGGATATAAACGATGTTTGCATATCTCTGGAGCTTAATCTCATCCTGTTTGTATTTCTGTTGTTGGCTTAGTGGTTCAGTTCATTTTGCTTTTGCTCTACAGTTCATTAGTTCAGGTAAACATACCAGTCTTGACAGCCATTTTGTATTATCCTTGCTTCTCTATTCTG'
}

seqs = {K: SeqRecord(Seq(V)) for K, V in seqs.items()}
"""Sequences of contigs used in these tests"""

known_juncs = {
    "YNL130C"   : ["YNL130C:0-53^145-180(-)",],
    "YPL249C-A" : ["YPL249C-A:0-53^291-334(-)",],

    'YBR215W_mRNA_0'  : ['YBR215W_mRNA_0:0-108^192-2175(+)'],
    'YHL001W_mRNA_0'  : ['YHL001W_mRNA_0:0-146^544-961(+)'],
    'YIL018W_mRNA_0'  : ['YIL018W_mRNA_0:0-30^430-1280(+)'],
    'YIL133C_mRNA_0'  : ['YIL133C_mRNA_0:0-648^938-1007(-)'],
    'YIL156W_B_mRNA_0': ['YIL156W_B_mRNA_0:0-41^103-408(+)'],
    'YKL006W_mRNA_0'  : ['YKL006W_mRNA_0:0-157^555-954(+)'],
    'YMR194C_B_mRNA_0': ['YMR194C_B_mRNA_0:0-325^397-729(-)'],
    'YNL130C_mRNA_0'  : ['YNL130C_mRNA_0:0-1204^1296-1382(-)'],
    'YPL249C_A_mRNA_0': ['YPL249C_A_mRNA_0:0-415^653-697(-)'],
}
known_juncs = { K : [SegmentChain.from_str(X) for X in V] for K,V in known_juncs.items() }
"""Annotated splice junctions"""

all_known_juncs = []
for v in known_juncs.values():
    all_known_juncs.extend(v)

known_juncs_as_tuples = {
    "YNL130C"   : [("YNL130C",53,145,"-"),],
    "YPL249C-A" : [("YPL249C-A",53,291,"-"),],

    'YBR215W_mRNA_0'  : [('YBR215W_mRNA_0',108,192,'+'),],
    'YHL001W_mRNA_0'  : [('YHL001W_mRNA_0',146,544,'+'),],
    'YIL018W_mRNA_0'  : [('YIL018W_mRNA_0',30,430,'+'),],
    'YIL133C_mRNA_0'  : [('YIL133C_mRNA_0',648,938,'-'),],
    'YIL156W_B_mRNA_0': [('YIL156W_B_mRNA_0',41,103,'+'),],
    'YKL006W_mRNA_0'  : [('YKL006W_mRNA_0',157,555,'+'),],
    'YMR194C_B_mRNA_0': [('YMR194C_B_mRNA_0',325,397,'-'),],
    'YNL130C_mRNA_0'  : [('YNL130C_mRNA_0',1204,1296,'-'),],
    'YPL249C_A_mRNA_0': [('YPL249C_A_mRNA_0',415,653,'-'),],
}

query_juncs = {
    "YNL130C": [
        "YNL130C:0-49^141-180(-)",  # slid within range
        "YNL130C:0-53^145-180(-)",  # exact match
    ],
    "YPL249C-A": [
        "YPL249C-A:0-53^291-334(-)",  # exact match
        "YPL249C-A:0-49^287-334(-)",  # slid in minus range
        "YPL249C-A:0-54^292-334(-)",  # slid in plus range
    ],
    'YBR215W_mRNA_0': [
        'YBR215W_mRNA_0:0-106^190-2175(+)', 'YBR215W_mRNA_0:0-107^191-2175(+)',
        'YBR215W_mRNA_0:0-108^192-2175(+)'
    ],
    'YHL001W_mRNA_0': [
        'YHL001W_mRNA_0:0-144^542-961(+)', 'YHL001W_mRNA_0:0-145^543-961(+)',
        'YHL001W_mRNA_0:0-146^544-961(+)', 'YHL001W_mRNA_0:0-147^545-961(+)',
        'YHL001W_mRNA_0:0-148^546-961(+)'
    ],
    'YIL018W_mRNA_0': [
        'YIL018W_mRNA_0:0-29^429-1280(+)', 'YIL018W_mRNA_0:0-30^430-1280(+)',
        'YIL018W_mRNA_0:0-31^431-1280(+)', 'YIL018W_mRNA_0:0-32^432-1280(+)',
        'YIL018W_mRNA_0:0-33^433-1280(+)'
    ],
    'YIL133C_mRNA_0': [
        'YIL133C_mRNA_0:0-645^935-1007(-)', 'YIL133C_mRNA_0:0-646^936-1007(-)',
        'YIL133C_mRNA_0:0-647^937-1007(-)', 'YIL133C_mRNA_0:0-648^938-1007(-)',
        'YIL133C_mRNA_0:0-649^939-1007(-)'
    ],
    'YIL156W_B_mRNA_0': [
        'YIL156W_B_mRNA_0:0-41^103-408(+)', 'YIL156W_B_mRNA_0:0-42^104-408(+)',
        'YIL156W_B_mRNA_0:0-43^105-408(+)', 'YIL156W_B_mRNA_0:0-44^106-408(+)'
    ],
    'YKL006W_mRNA_0': [
        'YKL006W_mRNA_0:0-155^553-954(+)', 'YKL006W_mRNA_0:0-156^554-954(+)',
        'YKL006W_mRNA_0:0-157^555-954(+)', 'YKL006W_mRNA_0:0-158^556-954(+)',
        'YKL006W_mRNA_0:0-159^557-954(+)'
    ],
    'YMR194C_B_mRNA_0': [
        'YMR194C_B_mRNA_0:0-325^397-729(-)', 'YMR194C_B_mRNA_0:0-326^398-729(-)',
        'YMR194C_B_mRNA_0:0-327^399-729(-)'
    ],
    'YPL249C_A_mRNA_0': [
        'YPL249C_A_mRNA_0:0-411^649-697(-)', 'YPL249C_A_mRNA_0:0-412^650-697(-)',
        'YPL249C_A_mRNA_0:0-413^651-697(-)', 'YPL249C_A_mRNA_0:0-414^652-697(-)',
        'YPL249C_A_mRNA_0:0-415^653-697(-)', 'YPL249C_A_mRNA_0:0-416^654-697(-)'
    ]
}

query_juncs = {K: [SegmentChain.from_str(X) for X in V] for K, V in query_juncs.items()}
"""Discovered splice junctions within match ranges of known/annotated splice junctions"""

noncan_juncs = {
    "YNL130C": [
        "YNL130C:0-49^141-180(-)",  # slid within range
    ],
    "YPL249C-A": [
        "YPL249C-A:0-49^287-334(-)",  # slid in minus range
        "YPL249C-A:0-54^292-334(-)",  # slid in plus range
    ],
}
noncan_juncs = {K: [SegmentChain.from_str(X) for X in V] for K, V in noncan_juncs.items()}
"""Non-canonical splice junctions within match ranges of known/annotated splice junctions"""

match_ranges = {
    "YNL130C"         : [(-4, 0)],
    "YPL249C-A"       : [(-4, 1)],
    'YBR215W_mRNA_0'  : [(-2, 0)],
    'YHL001W_mRNA_0'  : [(-2, 2)],
    'YIL018W_mRNA_0'  : [(-1, 3)],
    'YIL133C_mRNA_0'  : [(-3, 1)],
    'YIL156W_B_mRNA_0': [(0, 3)],
    'YKL006W_mRNA_0'  : [(-2, 2)],
    'YMR194C_B_mRNA_0': [(0, 2)],
    'YPL249C_A_mRNA_0': [(-4, 1)],
} # yapf: disable
"""Match ranges for junctions in ``known_juncs`` above"""

unmatched_query_juncs = {
    "YNL130C": [
        "YNL130C:0-23^145-180(-)",
        "YNL130C:0-53^165-180(-)",
        "YNL130C:0-70^141-180(-)",
        "YNL130C:0-49^121-180(-)",
        "YNL130C:0-53^145-180(+)",  # exact match wrong strand
        "YNL130C_fake:0-53^145-180(-)",  # exact match wrong chromosome
    ],
    # these below are all 1 nucleotide outside match range
    'YBR215W_mRNA_0': ['YBR215W_mRNA_0:0-105^189-2175(+)', 'YBR215W_mRNA_0:0-109^193-2175(+)'],
    'YHL001W_mRNA_0': ['YHL001W_mRNA_0:0-143^541-961(+)', 'YHL001W_mRNA_0:0-149^547-961(+)'],
    'YIL018W_mRNA_0': ['YIL018W_mRNA_0:0-28^428-1280(+)', 'YIL018W_mRNA_0:0-34^434-1280(+)'],
    'YIL133C_mRNA_0': ['YIL133C_mRNA_0:0-644^934-1007(-)', 'YIL133C_mRNA_0:0-650^940-1007(-)'],
    'YIL156W_B_mRNA_0': ['YIL156W_B_mRNA_0:0-40^102-408(+)', 'YIL156W_B_mRNA_0:0-45^107-408(+)'],
    'YKL006W_mRNA_0': ['YKL006W_mRNA_0:0-154^552-954(+)', 'YKL006W_mRNA_0:0-160^558-954(+)'],
    'YMR194C_B_mRNA_0': ['YMR194C_B_mRNA_0:0-324^396-729(-)', 'YMR194C_B_mRNA_0:0-328^400-729(-)'],
    'YPL249C_A_mRNA_0': ['YPL249C_A_mRNA_0:0-410^648-697(-)', 'YPL249C_A_mRNA_0:0-417^655-697(-)']
}
unmatched_query_juncs = {
    K: [SegmentChain.from_str(X) for X in V]
    for K, V in unmatched_query_juncs.items()
}
"""Query junctions with no known matches"""

unmatched_noncan_query_juncs = [
    "YNL130C:0-23^145-180(-)",
    "YNL130C:0-53^165-180(-)",
    "YNL130C:0-70^141-180(-)",
    "YNL130C:0-49^121-180(-)",
]
unmatched_noncan_query_juncs = [SegmentChain.from_str(X) for X in unmatched_noncan_query_juncs]
"""Query junctions without canonical splice junctions in the match range"""

repetitive_regions = [
    "YBR215W_mRNA_0:190-193(+)",  # threeprime splice site plus
    "YHL001W_mRNA_0:144-149(+)",  # fiveprime splice site plus
    "YIL133C_mRNA_0:935-940(-)",  # threeprime splice site minus
    "YMR194C_B_mRNA_0:325-328(-)",  # fiveprime splice site minus
]
cross_hash = GenomeHash([SegmentChain(GenomicSegment.from_str(X)) for X in repetitive_regions])
cross_hash_seqs = {X.chrom for X in cross_hash.feature_dict.values()}



#!/usr/bin/env python
"""Tools for reading values from binary files

See Also
--------
:py:mod:`struct`
    Binary data structures in Python
"""
import numpy
import sys
import struct
from collections import namedtuple


class BinaryParserFactory(object):
    """Parser factory for different types of binary records.
    
    Creates parsers that unpack binary byte streams into dictionaries
    that match field names to values. These parsers are most useful as components
    of binary file readers.

        
    Parameters
    ----------
    name : str
        Name for parser
    
    fmt : str
        String specifying binary format of data. See :py:mod:`struct`
    
    fields : list
        Ordered list of field names to bind to data unpacked from binary file
        
    
    Attributes
    ----------
    name : str
        Human-readable name for parser
    
    fmt : str
        String specifying binary format of data, as specified in :py:mod:`struct`
    
    fields : list
        List of strings specifying variable names to bind to data
        when unpacked from a binary file, in same order as items in ``fmt``
    
    nt : :class:`~collections.namedtuple`
        A :class:`~collections.namedtuple` instance that will provide names
        to the unpacked data

    
    Examples
    --------
    A binary RGB color parser::
    
        >>> ColorParser = BinaryParserFactory("ColorParser","3Q",["r","g","b"])
        >>> fh = open("some_binary_file_containing_colors.bin","rb") # 'b' is important in mode flag!!
        >>> fh.seek(byte_location_of_an_rgb_color)
        >>> rgb_dict = ColorParser(fh) # read and parse 3 8-bit integers from file
        >>> rgb_dict
            { "r" : 255,
              "g" : 0,
              "b" : 52 }


    See Also
    --------
    struct
        For information on format strings
    """

    def __init__(self, name, fmt, fields):
        """Create a |BinaryParserFactory|
        
        Parameters
        ----------
        name : str
            Name for parser
        
        fmt : str
            String specifying binary format of data. See :py:mod:`struct`
        
        fields : list
            Ordered list of field names to bind to data unpacked from binary file
        """
        self.name = name
        self.fmt = fmt
        self.fields = fields
        self.nt = namedtuple(name, fields)

    def __str__(self):
        return "<%s fmt='%s' fields='%s'>" % (self.name, self.fmt, ",".join(self.fields))

    def __repr__(self):
        return str(self)

    def __call__(self, fh, byte_order="<"):
        """Parse data from `fh` into a dictionary mapping field names to their values
        
        Parameters
        ----------
        fh : byte stream
            File-like pointing to binary data. Pointer in file must be
            aligned with start of record.
        
        byte_order : str
            Character indicating endian-ness of data (default: `'<'` for little-endian)
        
        Returns
        -------
        :py:class:`~collections.OrderedDict`
            Dictionary mapping field names from `self.fields` to their values
        """
        tmp_dict = self.nt._make(
            struct.unpack(byte_order + self.fmt, fh.read(self.calcsize(byte_order)))
        )._asdict()
        for k in tmp_dict:
            if isinstance(tmp_dict[k], bytes):
                # Python 3.x returns bytes
                # convert byte objects to strings
                tmp_dict[k] = tmp_dict[k].decode("ascii")
            if sys.version_info < (3, ) and isinstance(tmp_dict[k], unicode):
                # Python 2.x returns unicodes
                # convert unicode to strings
                tmp_dict[k] = str(tmp_dict[k].decode("ascii"))

        return tmp_dict

    def calcsize(self, byte_order="<"):
        """Return calculated size, in bytes, of record

        Parameters
        ----------
        byte_order : str
            Character indicating endian-ness of data (default: `'<'` for little-endian)

        Returns
        -------
        int
            Calculated size of record, in bytes
        """
        return struct.calcsize(byte_order + self.fmt)


def find_null_bytes(inp, null=b"\x00"):
    """Finds all null characters in a byte-formatted input string
    
    Parameters
    ----------
    inp : bytes (or str in Python 2.7)
        byte string
    
    Returns
    -------
    :py:class:`numpy.ndarray`
        numpy array of integers indexing where the null character *\x00* was found
    """
    indices = []
    last_found = inp.find(null)
    while last_found > -1:
        indices.append(last_found)
        last_found = inp.find(null, 1 + last_found)

    return numpy.array(indices)
